"""
in the same way as python has functools, the module also has a few functions
to help with the function manipulation.
"""

from .utils import call_base_str


def and_(*funcs):
    """
    Returns a function that performs an 'and' operation on multiple functions.
    """

    def _(*args, call_base=None, **kwargs):
        return all(
            call_base_str(func, *args, call_base=call_base, **kwargs) for func in funcs
        )

    return _


def or_(*funcs):
    """
    Returns a function that performs an 'or' operation on multiple functions.
    """

    def _(*args, call_base=None, **kwargs):
        return any(
            call_base_str(func, *args, call_base=call_base, **kwargs) for func in funcs
        )

    return _


def not_(func):
    """
    Returns a function that performs a 'not' operation on a function.
    """

    def _(*args, call_base=None, **kwargs):
        return not call_base_str(func, *args, call_base=call_base, **kwargs)

    return _



from typing import Any, Mapping, Optional, Union, Dict, Tuple, Iterable
from pprint import pformat
from decimal import Decimal

from django.core.exceptions import ValidationError
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from .validators import call_validator, PreviewValidator, PreviewValidationError
from . import PREVIEW, pre, TCallableStr
from ..utils import call_base_str

TNumber = Union[int, float, Decimal]


def mix(*cls):
    """
    Returns a mix of types. Mixins should go first and the last one should be the main type.

    Example:
    mix(HTMLMixin, SimpleInt)
    """
    return type("", cls, {})


class ProcessorsMixin:
    """
    Mixin that adds processors to the type.

    The processors is a pipeline of functions that are applied to the py object.
    """

    processors: Iterable[TCallableStr] = ()

    def get_processors(self):
        return self.processors

    def to_python(self, value: str) -> Any:
        py_value = super().to_python(value)
        for processor in self.get_processors():
            py_value = call_base_str(processor, py_value)
        return py_value


class GiveProcessorsMixin:
    """
    Mixin that adds processors to the type.
    """

    give_processors: Iterable[
        Union[TCallableStr, Tuple[Optional[str], TCallableStr]]
    ] = ()

    def get_give_processors(self):
        return self.give_processors

    def give(self, value: Any, suffix: Optional[str] = None):
        ret = super().give(value, suffix)

        for processor_pair in self.get_give_processors():
            if isinstance(processor_pair, tuple):
                processor_suffix, processor_func = processor_pair
            else:
                processor_suffix = None
                processor_func = processor_pair

            if processor_suffix == suffix:
                ret = call_base_str(processor_func, ret)

        return ret


class MinMaxValidationMixin:
    """
    Mixin that validates that value is between min_value and max_value.

    Attributes:
    min_value: Minimum value. If None, then no minimum value.
    max_value: Maximum value. If None, then no maximum value.
    """

    min_value: Optional[TNumber] = None
    max_value: Optional[TNumber] = None

    def validate(self, value):
        super().validate(value)

        is_none_valid = self.min_value is None and self.max_value is None
        if not is_none_valid and value is None:
            raise ValidationError(_("Value cannot be None"))

        if self.min_value is not None and value < self.min_value:
            raise ValidationError(
                _("Value cannot be less than %(min_value)s")
                % {"min_value": self.min_value}
            )

        if self.max_value is not None and value > self.max_value:
            raise ValidationError(
                _("Value cannot be greater than %(max_value)s")
                % {"max_value": self.max_value}
            )

    def get_help_format(self):
        yield from super().get_help_format()

        if self.min_value is not None:
            yield f" from {self.min_value}"

        if self.max_value is not None:
            yield f" to {self.max_value}"


class EmptyNoneMixin:
    """
    Mixin for types that returns None if value is empty string.

    Works only for `value_required=False`
    """

    def to_python(self, value):
        if not value and not self.value_required:
            return None
        return super().to_python(value)

    def validate_value(self, value: str):
        if not value and not self.value_required:
            return
        super().validate_value(value)


class HTMLMixin:
    """
    Mixin for types that should be displayed in HTML format.
    And also returned content should be marked as safe.
    """

    admin_preview_as: PREVIEW = PREVIEW.HTML

    def get_help_format(self):
        yield from super().get_help_format()
        yield " "
        yield _("in HTML format")

    def give(self, value, suffix=None):
        return mark_safe(super().give(value, suffix))


class PositiveValidationMixin(MinMaxValidationMixin):
    """
    Mixin that validates that value is positive.
    """

    min_value = 0


class CallToPythonMixin:
    """
    Mixin for callable types, or types that should be called to get the value.
    """

    call_func_argument_name: str = "prepared"
    call_func: TCallableStr = staticmethod(
        lambda *args, prepared=None, **kwargs: prepared(*args, **kwargs)
    )
    call_prepare_func: TCallableStr = staticmethod(lambda value: value)

    def get_call_prepare_func(self) -> TCallableStr:
        return self.call_prepare_func

    def prepare_python_call(self, value: Any) -> Dict[str, Any]:
        return {
            self.call_func_argument_name: call_base_str(
                self.get_call_prepare_func(), value
            )
        }

    def get_call_func(self) -> TCallableStr:
        return self.call_func

    def python_call(self, *args, **kwargs):
        return call_base_str(self.get_call_func(), *args, **kwargs)

    def get_preview_validators(self):
        return [validator for validator in self.get_validators()]

    def to_python(self, value):
        value = super().to_python(value)
        prepared_kwargs = self.prepare_python_call(value)

        def ret(*args, **kwargs):
            return self.python_call(*args, **prepared_kwargs, **kwargs)

        return ret

    def give_python_to_admin(self, value, name, **kwargs):
        try:
            value = self.to_python(value)
        except Exception as e:
            return [(None, e)]
        ret = []
        for validator in self.get_preview_validators():
            has_call_representation = getattr(
                validator, "has_call_representation", False
            )
            if has_call_representation:
                str_validator = f"{name}({str(validator)})"
            else:
                str_validator = f"{name}(???)"

            try:
                ret_validator = validator(value)
            except PreviewValidationError as e:
                ret.append((f"{name}({e.preview_input})", e))
            except Exception as e:
                ret.append((str_validator, e))
            else:
                if isinstance(ret_validator, PreviewValidator):
                    ret.append(
                        (
                            f"{name}({ret_validator.preview_input})",
                            ret_validator.preview_output,
                        )
                    )
                elif has_call_representation:
                    ret.append((str_validator, ret_validator))
        return ret

    def get_admin_preview_object(self, value, name, **kwargs):
        admin_preview_as = self.get_admin_preview_as()
        if admin_preview_as == PREVIEW.NONE:
            return ""

        if not value:
            return _("No preview (add at least one call_validator in validators)")

        if len(value) == 1 and admin_preview_as in (PREVIEW.TEXT, PREVIEW.HTML):
            value = value[0][1]
            if isinstance(value, Exception):
                return f"ERROR!!! {value}"
            if admin_preview_as == PREVIEW.TEXT:
                return pre(value)
            else:
                return str(value)

        def f():
            for validator, val in value:
                if validator is not None:
                    yield pre(f">>> {validator}")

                if isinstance(val, Exception):
                    yield f"ERROR!!! {val}"
                else:
                    if admin_preview_as == PREVIEW.TEXT:
                        yield pre(val)
                    elif admin_preview_as == PREVIEW.HTML:
                        yield str(val)
                    else:
                        yield pre("<<< " + pformat(val))

        return "\n".join(f())


class GiveCallMixin:
    """
    Mixin for callable types, but result of the call without artuments should be returned.

    If suffix is "call" then callable should be returned.
    """

    def get_suffixes_names(self, value: Any):
        return ("call",) + super().get_suffixes_names(value)

    def get_validators(self):
        return (call_validator(),) + super().get_validators()

    def get_validators(self):
        validators = super().get_validators()
        if not validators:
            return (call_validator(),)
        return validators

    def give(self, value, suffix=None):
        if suffix is None:
            return value()
        if suffix == "call":
            return value
        return super().give(value, suffix)


class MakeCallMixin:
    """
    Mixin for non-callable python objects will be returned as callable given.

    Can be usefull when you change callable types to a simple type but don't want to change the code that uses that type.
    """

    def get_suffixes_names(self, value: Any):
        return ("call",) + super().get_suffixes_names(value)

    def give_python_to_admin(self, value, name, **kwargs):
        return self.give_python(value)()

    def give(self, value, suffix=None):
        return lambda *args, **kwargs: value


class AdminPreviewMenuMixin:
    """
    Mixin that adds a menu to the admin preview.
    """

    def gen_admin_preview_html_menu_items(self, value: Any, name: str, **kwargs) -> str:
        return
        yield

    def get_admin_preview_html_menu(self, value: Any, name: str, **kwargs) -> str:
        html_items = "".join(
            self.gen_admin_preview_html_menu_items(value, name, **kwargs)
        )
        if not html_items:
            return ""
        return "<div>" + html_items + "</div>"

    def get_admin_preview_object(self, value: Any, name: str, **kwargs) -> str:

        return self.get_admin_preview_html_menu(
            value, name, **kwargs
        ) + self.get_admin_preview_under_menu_object(value, name, **kwargs)

    def get_admin_preview_under_menu_object(
        self, value: Any, name: str, **kwargs
    ) -> str:
        return super().get_admin_preview_object(value, name, **kwargs)


class AdminSuffixesMixinPreview:
    admin_preview_suffixes = None
    admin_preview_suffixes_default = "default"

    def get_admin_preview_suffixes_default(self):
        return self.admin_preview_suffixes_default

    def get_admin_preview_suffixes(self, value: Any, name: str, **kwargs):
        return (
            self.get_suffixes_names(value)
            if self.admin_preview_suffixes is None
            else self.admin_preview_suffixes
        )

    def gen_admin_preview_html_menu_items(
        self, value: Any, name: str, suffix: Optional[str] = None, **kwargs
    ) -> str:
        suffixes = self.get_admin_preview_suffixes(value, name, **kwargs)
        if not len(suffixes):
            return

        for suf in (None, *suffixes):
            if suf == suffix:
                yield f" <b>{suf or self.get_admin_preview_suffixes_default()}</b> "
            elif suf is None:
                yield f' <a class="cs_set_params">{self.get_admin_preview_suffixes_default()}</a> '
            else:
                yield f' <a class="cs_set_params" data-param-suffix="{suf}">{suf}</a> '

        yield from super().gen_admin_preview_html_menu_items(
            value, name, suffix=suffix, **kwargs
        )

    def get_admin_preview_under_menu_object(
        self, value: Any, name: str, **kwargs
    ) -> str:
        suffix = kwargs.get("suffix", None)
        return super().get_admin_preview_under_menu_object(
            self.give(value, suffix), name, **kwargs
        )


class AdminPreviewSuffixesMixin(AdminSuffixesMixinPreview, AdminPreviewMenuMixin):
    """
    Mixin shows links to preview different suffixes of the value in the admin preview.
    """

    pass


class ActionResponse(dict):
    def error(self, message):
        self["error"] = message
        return self

    def html(self, html):
        self["html"] = html
        return self

    def alert(self, message):
        self["alert"] = message
        return self

    def value(self, value):
        self["value"] = value
        return self

    def before_html(self, html):
        self["before_html"] = html
        return self


class AdminActionsMixinPreview:
    """
    Mixin that adds actions to the admin preview.
    """

    admin_preview_actions = None

    def gen_admin_preview_html_menu_items(
        self, value: Any, name: str, suffix: Optional[str] = None, **kwargs
    ) -> str:
        if self.admin_preview_actions is None:
            return

        for i, (action_name, _) in enumerate(self.admin_preview_actions):
            yield f' <a class="cs_set_params cs_set_params_once" data-param-action="{i}">{action_name}</a> '

    def get_admin_preview_function(self, *args, **kwargs):
        if self.admin_preview_actions is None:
            return

        action_name = kwargs.get("action", None)
        if action_name is None:
            return

        try:
            return self.admin_preview_actions[int(action_name)][1]
        except (IndexError, ValueError):
            return

    def process_action(self, *args, **kwargs):
        func = self.get_admin_preview_function(*args, **kwargs)
        if func is None:
            return

        response = ActionResponse()
        func(response, *args, _cs_type=self, **kwargs)
        return response

    def get_full_admin_preview_value(self, *args, **kwargs) -> str:
        response = self.process_action(*args, **kwargs)
        if not response:
            return super().get_full_admin_preview_value(*args, **kwargs)

        response = dict(response)

        before_html = response.pop("before_html", "")
        if before_html:
            response = {
                **response,
                **super().get_full_admin_preview_value(*args, **kwargs),
            }
            response["html"] = before_html + response["html"]

        return response


class AdminPreviewActionsMixin(AdminActionsMixinPreview, AdminPreviewMenuMixin):
    pass



# update docs/settings.md

from django.conf import settings

from . import __version__


def get_setting(name, default=None):
    return getattr(settings, "CONTENT_SETTINGS_" + name, default)


PREVIEW_ON_SITE_HREF = get_setting("PREVIEW_ON_SITE_HREF", "/")

PREVIEW_ON_SITE_SHOW = get_setting("PREVIEW_ON_SITE_SHOW", True)

UPDATE_DB_VALUES_BY_MIGRATE = get_setting("UPDATE_DB_VALUES_BY_MIGRATE", True)

TAGS = get_setting(
    "TAGS",
    [],
)

TAG_CHANGED = get_setting("TAG_CHANGED", "changed")

CHECKSUM_KEY_PREFIX = (
    get_setting("CHECKSUM_KEY_PREFIX", "CS_CHECKSUM_") + __version__ + "__"
)

CHECKSUM_USER_KEY_PREFIX = (
    get_setting("CHECKSUM_KEY_PREFIX", "CS_USER_CHECKSUM_") + __version__ + "__"
)

USER_TAGS = get_setting(
    "USER_TAGS",
    {
        "favorites": ("⭐", "⚝"),
        "marked": ("💚", "♡"),
    },
)

CACHE_TRIGGER = get_setting(
    "CACHE_TRIGGER", "content_settings.cache_triggers.VersionChecksum"
)
if isinstance(CACHE_TRIGGER, str):
    CACHE_TRIGGER = {
        "backend": CACHE_TRIGGER,
    }
elif isinstance(CACHE_TRIGGER, dict) and "backend" not in CACHE_TRIGGER:
    CACHE_TRIGGER = {
        "backend": "content_settings.cache_triggers.VersionChecksum",
        **CACHE_TRIGGER,
    }

VALUES_ONLY_FROM_DB = get_setting("VALUES_ONLY_FROM_DB", False) and not settings.DEBUG

VALIDATE_DEFAULT_VALUE = get_setting("VALIDATE_DEFAULT_VALUE", settings.DEBUG)

DEFAULTS = get_setting("DEFAULTS", [])

ADMIN_CHECKSUM_CHECK_BEFORE_SAVE = get_setting(
    "ADMIN_CHECKSUM_CHECK_BEFORE_SAVE", False
)

CHAIN_VALIDATE = get_setting("CHAIN_VALIDATE", True)

UI_DOC_URL = get_setting(
    "UI_DOC_URL", "https://django-content-settings.readthedocs.io/en/0.25/ui/"
)

CHECK_UPDATE_CELERY = get_setting("CHECK_UPDATE_CELERY", True)

CHECK_UPDATE_HUEY = get_setting("CHECK_UPDATE_HUEY", True)

USER_DEFINED_TYPES = get_setting("USER_DEFINED_TYPES", [])

assert isinstance(
    USER_DEFINED_TYPES, list
), "CONTENT_SETTINGS_USER_DEFINED_TYPES must be a list"
assert len(USER_DEFINED_TYPES) == len(
    set([val[0] for val in USER_DEFINED_TYPES])
), "CONTENT_SETTINGS_USER_DEFINED_TYPES must have unique slugs"
for i, val in enumerate(USER_DEFINED_TYPES):
    assert (
        len(val) == 3
    ), f"CONTENT_SETTINGS_USER_DEFINED_TYPES[{i}] must be a tuple of 3 elements"
    assert all(
        [isinstance(v, str) for v in val]
    ), f"CONTENT_SETTINGS_USER_DEFINED_TYPES[{i}] must be a tuple of 3 strings"

    slug, imp_line, name = val
    assert (
        len(slug) < 50
    ), f"CONTENT_SETTINGS_USER_DEFINED_TYPES[{i}][0] must be less than 50 chars"



"""
    Modifiers are functions that take kwargs and return updates for that dict.

    Modifiers are using in `DEFAULTS` second element of the tuple and as `defaults` arguments.
"""

from typing import Dict, Any, Callable, Set, Iterable, Tuple, Optional

TModifier = Callable[[Dict[str, Any]], Dict[str, Any]]


class NotSet:
    """
    A reference to say that the value is not set. (Using None is not possible)
    """

    pass


class SkipSet(Exception):
    """
    For unite modifiers, to skip setting the value.
    """


def set_if_missing(**params: Any) -> TModifier:
    """
    Set key-value pairs in the updates dictionary if they are not already set in kwargs. This modifier is used for all `**kwargs` attributes for `defaults` context.

    Args:
        **params: Arbitrary keyword arguments representing key-value pairs to set.
    """
    return lambda type_kwargs, updates, kwargs: {
        k: v if k not in kwargs else NotSet for k, v in params.items()
    }


class unite(object):
    """
    unite is a base class for modifiers that unites kwargs passed in arguments with kwargs already collected and kwargs passed in the definition of the settings type.

    All child classes should implement `process` method.

    Args:
        * **kwargs: Arbitrary keyword arguments representing key-value pairs to unite.
        _use_type_kwargs: (default True) A boolean indicating whether to use type kwargs.
        _empty_not_set: (default False) A boolean indicating whether empty value should be removed from updates.
    """

    use_type_kwargs: bool = True
    empty_not_set: bool = False

    def __init__(self, **kwargs) -> None:
        self.use_type_kwargs = kwargs.pop("_use_type_kwargs", self.use_type_kwargs)
        self.empty_not_set = kwargs.pop("_empty_not_set", self.empty_not_set)
        self.params = kwargs

    def __call__(
        self,
        type_kwargs: Dict[str, Any],
        updates: Dict[str, Any],
        kwargs: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        Unites the provided updates and kwargs dictionaries with the parameters.

        Args:
            type_kwargs: The default kwargs of the settings type.
            updates: The dictionary with already collected kwargs from default context.
            kwargs: The kwargs passed in the definition of the settings type.

        Returns:
            A dictionary with united key-value pairs.
        """
        result = {}
        for k, v in self.params.items():
            if k not in type_kwargs:
                continue
            try:
                process_result = self.process(
                    v,
                    type_kwargs[k] if self.use_type_kwargs else NotSet,
                    updates.get(k, NotSet),
                    kwargs.get(k, NotSet),
                )
            except SkipSet:
                continue
            else:
                result[k] = (
                    NotSet
                    if self.empty_not_set and not process_result
                    else process_result
                )
        return result

    def process(self, value: Any, tw: Any, up: Any, kw: Any) -> Any:
        """
        Returns value for the update dictionary.

        Args:
            value: The value from the parameter of the modifier.
            tw: The current value in the type kwargs.
            up: The current value in the update dictionary.
            kw: The current value in the settings kwargs.

        Returns:
            The processed value to be set in the update dictionary.
        """
        raise NotImplementedError("Subclass must implement process method")


class unite_set_add(unite):
    """
    unite that modifies a set by adding new values in it.
    """

    def process(
        self,
        value: Iterable[Any],
        tw: Optional[Iterable[Any]],
        up: Optional[Iterable[Any]],
        kw: Optional[Iterable[Any]],
    ) -> Set[Any]:
        return (
            set(value)
            | (set(up) if up and up is not NotSet else set())
            | (set(kw) if kw and kw is not NotSet else set())
            | (set(tw) if tw and tw is not NotSet else set())
        )


class unite_set_remove(unite):
    """
    unite that modifies a set by removing given values.

    """

    empty_not_set: bool = True

    def process(
        self,
        value: Iterable[Any],
        tw: Optional[Iterable[Any]],
        up: Optional[Iterable[Any]],
        kw: Optional[Iterable[Any]],
    ) -> Set[Any]:
        if up is NotSet or not up:
            raise SkipSet()

        return set(up) - set(value) | (set(kw) if kw and kw is not NotSet else set())


class unite_tuple_add(unite):
    """
    unite that modifies a tuple by adding new values in it.
    """

    def process(
        self,
        value: Iterable[Any],
        tw: Optional[Iterable[Any]],
        up: Optional[Iterable[Any]],
        kw: Optional[Iterable[Any]],
    ) -> Tuple[Any]:
        if up and up is not NotSet:
            old_value = up
        elif kw and kw is not NotSet:
            old_value = kw
            if tw and tw is not NotSet:
                old_value = tuple(old_value) + tuple(tw)
        elif tw and tw is not NotSet:
            old_value = tw
        else:
            old_value = ()
        return tuple(value) + tuple(old_value)


class unite_dict_update(unite):
    """
    unite that modifies a dictionary by updating it with new values.
    """

    def process(
        self,
        value: Dict[str, Any],
        tw: Optional[Dict[str, Any]],
        up: Optional[Dict[str, Any]],
        kw: Optional[Dict[str, Any]],
    ) -> Dict[str, Any]:
        return {
            **(up if up and up is not NotSet else {}),
            **value,
            **(kw if kw and kw is not NotSet else {}),
            **(tw if tw and tw is not NotSet else {}),
        }


def add_tags(tags: Iterable[str], **kwargs) -> TModifier:
    """
    add tags to the setting
    """
    return unite_set_add(tags=tags, **kwargs)


def remove_tags(tags: Iterable[str], **kwargs) -> TModifier:
    """
    Removes tags from the update context.
    """
    return unite_set_remove(tags=tags, **kwargs)


def add_tag(tag: str, **kwargs) -> TModifier:
    """
    same as `add_tags` but only one.
    """
    return unite_set_add(tags={tag}, **kwargs)


def remove_tag(tag: str, **kwargs) -> TModifier:
    """
    same as `remove_tags` but only one.
    """
    return unite_set_remove(tags={tag}, **kwargs)


class unite_str(unite):
    """
    unite that modifies a string by formatting it.

    Args:
        _format: A string to format the value use `{new_value}` as a placeholder for the passed value in kwargs and `{old_value}` as a placeholder for the value in the update context.
        **kwargs: Arbitrary keyword arguments.
    """

    def __init__(self, _format: str = "{new_value}\n{old_value}", **kwargs) -> None:
        self._format = _format
        super().__init__(**kwargs)

    def process(
        self, value: Any, tw: Optional[Any], up: Optional[Any], kw: Optional[Any]
    ) -> str:
        if up and up is not NotSet:
            old_value = up
        elif kw and kw is not NotSet:
            old_value = kw
            if tw and tw is not NotSet:
                old_value = self._format.format(new_value=kw, old_value=tw)
        elif tw and tw is not NotSet:
            old_value = tw
        else:
            old_value = ""

        return self._format.format(new_value=value, old_value=old_value)


def help_prefix(prefix: str) -> TModifier:
    """
    add prefix to help
    """
    return unite_str("{new_value}<br>{old_value}", help=prefix)


def help_suffix(suffix: str) -> TModifier:
    """
    add suffix to help
    """
    return unite_str("{old_value}<br>{new_value}", help=suffix)


def add_admin_head(
    css: Iterable[str] = (),
    js: Iterable[str] = (),
    css_raw: Iterable[str] = (),
    js_raw: Iterable[str] = (),
) -> TModifier:
    """
    add css and js to the admin head

    * css -> admin_head_css
    * js -> admin_head_js
    * css_raw -> admin_head_css_raw
    * js_raw -> admin_head_js_raw
    """
    return unite_tuple_add(
        **({"admin_head_css": css} if css else {}),
        **({"admin_head_js": js} if js else {}),
        **({"admin_head_css_raw": css_raw} if css_raw else {}),
        **({"admin_head_js_raw": js_raw} if js_raw else {}),
    )


class add_widget_class(unite_dict_update):
    """
    add widget class or classes splited by space.
    """

    def __init__(self, class_name: str):
        super().__init__(widget_attrs={"class": class_name})

    def process(
        self,
        value: Dict[str, Any],
        tw: Optional[Dict[str, Any]],
        up: Optional[Dict[str, Any]],
        kw: Optional[Dict[str, Any]],
    ) -> Dict[str, Any]:
        ret = super().process(value, tw, up, kw)
        if up is NotSet and kw is NotSet:
            return ret

        ret["class"] = " ".join(
            set(value["class"].split())
            | (
                set(up["class"].split())
                if up and up is not NotSet and "class" in up
                else set()
            )
            | (
                set(kw["class"].split())
                if kw and kw is not NotSet and "class" in kw
                else set()
            )
            | (
                set(tw["class"].split())
                if tw and tw is not NotSet and "class" in tw
                else set()
            )
        )
        return ret


def update_widget_attrs(**kwargs) -> TModifier:
    """
    update widget attrs
    """
    return unite_dict_update(widget_attrs=kwargs)



"""
A list of functions that are used as values for type attributes such as `fetch_permission`, `view_permission`, `update_permission`, and `view_history_permission`.
"""


def any(user):
    """
    Returns True for any user.
    """
    return True


def none(user):
    """
    Returns False for any user.
    """
    return False


def authenticated(user):
    """
    Returns True if the user is authenticated.
    """
    return user.is_authenticated


def staff(user):
    """
    Returns True if the user is active and a staff member.
    """
    return user.is_active and user.is_staff


def superuser(user):
    """
    Returns True if the user is active and a superuser.
    """
    return user.is_active and user.is_superuser


def has_perm(perm):
    """
    Returns a function that checks if the user has a specific permission.
    """

    def _(user):
        return user.is_active and user.has_perm(perm)

    return _



"""
Types that convert a string into a datetime, date, time or timedelta object.
"""

from typing import Any, List, Tuple, Union

from datetime import datetime, time, timedelta
from django.core.exceptions import ValidationError
from django.utils import formats as django_formats
from django.forms.utils import from_current_timezone
from django.utils.translation import gettext as _

from .basic import SimpleString, PREVIEW
from .mixins import EmptyNoneMixin

TIMEDELTA_FORMATS = {
    "s": "seconds",
    "m": "minutes",
    "h": "hours",
    "d": "days",
    "w": "weeks",
}


def timedelta_format(text: str) -> timedelta:
    """
    Convert a string into a timedelta object using format from `TIMEDELTA_FORMATS`.
    For example:
    - `1d` - one day
    - `1d 3h` - one day and three hours
    """
    text = text.strip()
    delta_kwargs = {}
    for el in text.split():
        try:
            val = int(el[:-1])
            val_key = el[-1]
            delta_kwargs[TIMEDELTA_FORMATS[val_key]] = val
        except Exception as e:
            raise ValidationError(
                _("wrong format for %(el)s: %(e)s") % {"el": el, "e": e}
            )

    return timedelta(**delta_kwargs)


class ProcessInputFormats(EmptyNoneMixin, SimpleString):
    """
    Base class for converting a string into a datetime, date or time object. (Teachnically can be used for other fields with predefined formats by overriding `postprocess_input_format` method.)

    It uses the `input_formats_field` from the Meta class to get the filed with formats.

    We want to use different field for different formats as we want to be able to override specific format in using `CONTENT_SETTINGS_CONTEXT`
    """

    admin_preview_as = PREVIEW.PYTHON

    class Meta:
        input_formats_field = None

    def postprocess_input_format(self, value: Any, format: Any) -> Any:
        """
        converts a given value using a given format
        """
        raise NotImplementedError

    def get_input_formats(self):
        assert (
            self.Meta.input_formats_field
        ), "You must define input_formats_field in Meta class"

        formats = getattr(self, self.Meta.input_formats_field)
        if isinstance(formats, (list, tuple)):
            return formats
        return [formats]

    def get_help_format(self):
        yield _("Use any of the following formats:")
        yield (
            " <ul>"
            + "".join([f"<li>{l}</li>" for l in self.get_input_formats()])
            + "</ul>"
        )

    def to_python(self, value):
        value = super().to_python(value)
        if value is None:
            return None

        for format in self.get_input_formats():
            try:
                return self.postprocess_input_format(value, format)
            except (ValueError, TypeError):
                continue
        raise ValidationError(_("Wrong Format"))


TFormats = Union[List[str], Tuple[str, ...], str]


class DateTimeString(ProcessInputFormats):
    """
    Converts into a datetime object.

    Attributes:

    - `datetime_formats` - list (or a single string) of formats to use for conversion. As a default it uses `DATETIME_INPUT_FORMATS` from `django.conf.settings`
    """

    datetime_formats: TFormats = list(
        django_formats.get_format_lazy("DATETIME_INPUT_FORMATS")
    )

    class Meta:
        input_formats_field = "datetime_formats"

    def postprocess_input_format(self, value, format):
        return from_current_timezone(datetime.strptime(value.strip(), format))


class DateString(ProcessInputFormats):
    """
    Converts into a date object.

    Attributes:

    - `date_formats` - list (or a single string) of formats to use for conversion. As a default it uses `DATE_INPUT_FORMATS` from `django.conf.settings`
    """

    date_formats: TFormats = list(django_formats.get_format_lazy("DATE_INPUT_FORMATS"))

    class Meta:
        input_formats_field = "date_formats"

    def postprocess_input_format(self, value, format):
        return datetime.strptime(value.strip(), format).date()


class TimeString(DateTimeString):
    """
    Converts into a time object.

    Attributes:

    - `time_formats` - list (or a single string) of formats to use for conversion. As a default it uses `TIME_INPUT_FORMATS` from `django.conf.settings`
    """

    time_formats: TFormats = list(django_formats.get_format_lazy("TIME_INPUT_FORMATS"))

    class Meta:
        input_formats_field = "time_formats"

    def postprocess_input_format(self, value, format):
        value = datetime.strptime(value.strip(), format)
        return time(
            value.hour, value.minute, value.second, value.microsecond, fold=value.fold
        )


class SimpleTimedelta(EmptyNoneMixin, SimpleString):
    """
    Converts into a timedelta object.
    """

    admin_preview_as = PREVIEW.PYTHON

    def get_help_format(self):
        yield _("Time Delta Format")
        yield "<ul>"
        for name, value in TIMEDELTA_FORMATS.items():
            yield f"<li>{name} - {value}</li>"
        yield "</ul>"
        yield _("Examples:")
        yield "<ul>"
        yield "<li>"
        yield _("1d - in one day")
        yield "</li>"
        yield "<li>"
        yield _("1d 3h - in one day and 3 hours")
        yield "</li>"
        yield "</ul>"

    def to_python(self, value):
        py_value = super().to_python(value)
        if py_value is None:
            return None
        return timedelta_format(py_value)



"""
Functions that can be used as filters for the `DEFAULTS` setting.

Each function has a single attribute *settings type* and should return a boolean.
"""

from typing import Callable, Type

from content_settings.types import BaseSetting
from content_settings.utils import class_names

TFilter = Callable[[Type[BaseSetting]], bool]


def any_name(cls: Type[BaseSetting]) -> bool:
    """
    Allow all settings.
    """
    return True


def name_exact(name: str) -> TFilter:
    """
    Allow only settings with the exact type name or parent type name.

    Args:
        name (str): The exact name to match.
    """

    def f(cls: Type[BaseSetting]) -> bool:
        return any(name == el[1] for el in class_names(cls))

    return f


def full_name_exact(name: str) -> TFilter:
    """
    Allow only settings with the exact full type name or parent type name. The name includes module name.

    Args:
        name (str): The exact full name to match, including the module.

    """

    def f(cls: Type[BaseSetting]) -> bool:
        return any(name == f"{el[0]}.{el[1]}" for el in class_names(cls))

    return f



"""
the caching backend is working with local thread storage to store the checksum raw and py objects.

`DATA` is a local thread storage with the following attributes:

* `ALL_RAW_VALUES: Dict[str, str]` - the raw values (values from the database) of the all settings
* `ALL_VALUES: Dict[str, Any]` - the python objects of the all settings
* `ALL_USER_DEFINES: Dict[str, BaseSetting]` - key is the setting name, value is the user defined type (with tags and help text)
* `POPULATED: bool` - the flag that indicates that all values were populated from the database
"""

import threading
from typing import Any, Dict, Set, Optional, List

from django.conf import settings

from .utils import import_object
from .types import BaseSetting
from .settings import (
    VALUES_ONLY_FROM_DB,
    VALIDATE_DEFAULT_VALUE,
    CACHE_TRIGGER,
    USER_DEFINED_TYPES,
)


TRIGGER = import_object(CACHE_TRIGGER["backend"])(
    **{k: v for k, v in CACHE_TRIGGER.items() if k != "backend"}
)


class ThreadLocalData(threading.local):
    POPULATED: bool = False
    ALL_RAW_VALUES: Optional[Dict[str, str]] = None
    ALL_VALUES: Optional[Dict[str, Any]] = None
    ALL_USER_DEFINES: Optional[Dict[str, BaseSetting]] = None


DATA = ThreadLocalData()


def get_form_checksum():
    if not is_populated():
        populate()
    return TRIGGER.get_form_checksum()


def set_new_type(
    name: str,
    user_defined_type: str,
    tags_set: Optional[Set[str]] = None,
    help: str = "",
) -> Optional[BaseSetting]:
    """
    create a new user defined type and saves it to the local thread. The previous type is returned.
    """
    if tags_set is None:
        tags_set = set()

    from .conf import USER_DEFINED_TYPES_INSTANCE

    if not is_populated():
        populate()

    prev_cs_type = cs_type = DATA.ALL_USER_DEFINES.get(name)

    if not cs_type or cs_type.tags != tags_set or cs_type.help != help:
        cs_type = USER_DEFINED_TYPES_INSTANCE[user_defined_type](
            help=help,
            tags=tags_set,
        )

    DATA.ALL_USER_DEFINES[name] = cs_type
    return prev_cs_type


def replace_user_type(name: str, cs_type: BaseSetting) -> Optional[BaseSetting]:
    """
    replace the user defined type with the new one. The previous type is returned.
    """
    if not is_populated():
        populate()

    prev_cs_type = DATA.ALL_USER_DEFINES.get(name)
    DATA.ALL_USER_DEFINES[name] = cs_type
    return prev_cs_type


def set_new_value(name: str, new_value: str, version: Optional[str] = None) -> str:
    """
    takes name, raw value and saves it to the local thread if the value was changed. The previous value is returned.

    raw value converts to the python object and saves to the local thread.

    if version is not None - it will be verified against the version of the type
    """
    cs_type = get_type_by_name(name)
    assert cs_type is not None, f"Can't find type for {name}"

    prev_value = get_raw_value(name)

    if version is None or cs_type.version == version and prev_value != new_value:
        DATA.ALL_RAW_VALUES[name] = new_value
        if name in DATA.ALL_VALUES:
            DATA.ALL_VALUES.pop(name)

    return prev_value


def set_new_db_value(name: str, value: str, *type_define) -> str:
    """
    set the new value for the setting in DB
    """
    from .models import ContentSetting
    from .conf import get_str_tags

    if type_define:
        assert USER_DEFINED_TYPES, "User defined types are not enabled"

    cs_type = get_type_by_name(name)
    assert (
        cs_type is not None or type_define
    ), f"{name} is not defined in any content_settings.py file"
    if not cs_type:
        set_new_type(name, *type_define)
        cs_type = get_type_by_name(name)

    cs_type.validate_value(value)

    try:
        cs = ContentSetting.objects.get(name=name)
    except ContentSetting.DoesNotExist:
        ContentSetting.objects.create(
            name=name,
            value=value,
            version=cs_type.version,
            tags=get_str_tags(name, cs_type, value),
            help=cs_type.get_help(),
            user_defined_type=cs_type.user_defined_slug,
        )
    else:
        assert cs.version == cs_type.version, "Version mismatch"

        cs.value = value
        cs.save()

    return set_new_value(name, value, version=None)


def delete_user_value(name: str) -> Optional[str]:
    """
    delete user defined setting from the local thread and returns its raw value
    """
    if get_raw_value(name) is None:
        return None

    prev_value = get_raw_value(name)
    del DATA.ALL_RAW_VALUES[name]
    if name in DATA.ALL_VALUES:
        del DATA.ALL_VALUES[name]
    del DATA.ALL_USER_DEFINES[name]

    return prev_value


def get_type_by_name(name: str) -> Optional[BaseSetting]:
    """
    get the type of the setting (inluding user defined types) by its name
    """
    from .conf import ALL

    if name in ALL:
        return ALL[name]

    if not USER_DEFINED_TYPES:
        return None

    return get_userdefined_type_by_name(name)


def get_userdefined_type_by_name(name: str) -> Optional[BaseSetting]:
    """
    get the user defined type by its name
    """
    if not is_populated():
        populate()

    return DATA.ALL_USER_DEFINES.get(name)


def get_value(name: str, suffix: Optional[str] = None) -> Any:
    """
    get the value of the setting by its name and optional suffix
    """

    cs_type = get_type_by_name(name)
    if cs_type is None:
        raise AttributeError(f"{name} is not defined in any content_settings.py file")

    return cs_type.give(get_py_value(name), suffix)


def get_raw_value(name: str) -> Optional[str]:
    """
    get the raw value of the setting by its name
    """
    from .conf import is_constant

    if is_constant(name):
        return get_type_by_name(name).default

    if not is_populated():
        populate()

    return DATA.ALL_RAW_VALUES.get(name)


def get_constant_py_value(name: str) -> Any:
    """
    get the python object of the constant setting by its name
    """
    # constant can work without populated data
    if DATA.ALL_VALUES and name in DATA.ALL_VALUES:
        return DATA.ALL_VALUES[name]

    cs_type = get_type_by_name(name)

    if DATA.ALL_VALUES is None:
        DATA.ALL_VALUES = {}
    DATA.ALL_VALUES[name] = cs_type.to_python(cs_type.default)

    return DATA.ALL_VALUES[name]


def get_py_value(name: str) -> Any:
    """
    get the python object of the setting by its name
    """
    from .conf import is_constant

    if is_constant(name):
        return get_constant_py_value(name)

    if not is_populated():
        populate()

    if name in DATA.ALL_VALUES:
        return DATA.ALL_VALUES[name]

    assert name in DATA.ALL_RAW_VALUES, f"{name} is unknown"

    cs_type = get_type_by_name(name)

    DATA.ALL_VALUES[name] = cs_type.to_python(get_raw_value(name))

    return DATA.ALL_VALUES[name]


def is_populated() -> bool:
    """
    check if the local thread is populated with the values from the database and it is actual
    """
    return getattr(DATA, "POPULATED", False)


def set_populated(value: bool = True) -> None:
    DATA.POPULATED = value


def get_db_objects() -> Dict[str, Any]:
    """
    get the database objects for the settings
    """
    from .models import ContentSetting

    return {v.name: v for v in ContentSetting.objects.all()}


def get_all_names() -> List[str]:
    """
    get the names of the settings (including user defined types) from the local thread
    """
    if not is_populated():
        populate()

    return list(DATA.ALL_RAW_VALUES.keys())


def populate() -> None:
    """
    reset the local thread with the values from the database
    """
    if is_populated():
        return

    set_populated(True)
    if DATA.ALL_RAW_VALUES is None:
        DATA.ALL_VALUES = {}
        DATA.ALL_RAW_VALUES = {}
        DATA.ALL_USER_DEFINES = {}

    # test DB access
    try:
        from .models import ContentSetting

        ContentSetting.objects.all().first()
    except Exception:
        return

    reset_values()
    TRIGGER.reset()


def validate_default_values():
    """
    validate default values for all of the registered settings.

    Only is VALIDATE_DEFAULT_VALUE is True.
    """

    if not VALIDATE_DEFAULT_VALUE:
        return

    from .conf import ALL

    for name, cs_type in ALL.items():
        if not isinstance(cs_type.default, str):
            continue
        try:
            cs_type.validate_value(cs_type.default)
        except Exception as e:
            raise AssertionError(f"Error validating {name}: {e}")


def reset_user_values(db: Optional[Dict[str, Any]] = None) -> None:
    """
    reset the local thread with the values from the database for user defined types

    if trigger_checksum is not the same as the checksum in the local thread, the checksum in the cache backend will be updated
    """
    from .conf import USER_DEFINED_TYPES_INSTANCE

    if db is None:
        db = get_db_objects()

    names = set()
    for name, cs in db.items():
        if not cs.user_defined_type:
            continue

        assert (
            cs.user_defined_type in USER_DEFINED_TYPES_INSTANCE
        ), f"{cs.user_defined_type} is not in USER_DEFINED_TYPES"
        names.add(name)
        set_new_type(
            name,
            cs.user_defined_type,
            cs.tags_set,
            cs.help,
        )
        set_new_value(
            name,
            db[name].value,
            version=(None if settings.DEBUG else db[name].version),
        )

    for name in set(DATA.ALL_USER_DEFINES.keys()) - names:
        delete_user_value(name)


def reset_values(db: Optional[Dict[str, Any]] = None) -> None:
    """
    reset the local thread with the values from the database for code settings

    if trigger_checksum is not the same as the checksum in the local thread, the checksum in the cache backend will be updated
    """
    from .conf import ALL

    if db is None:
        db = get_db_objects()

    # the first run (not raw values)
    is_init = not bool(DATA.ALL_RAW_VALUES)

    for name, cs_type in ALL.items():
        if cs_type.constant:
            set_new_value(name, ALL[name].default)

        elif name in db:
            assert (
                not db[name].user_defined_type or cs_type.overwrite_user_defined
            ), f"{name} is not a code setting and not overwrite_user_defined"

            set_new_value(
                name,
                db[name].value,
                version=(None if settings.DEBUG else db[name].version),
            )

        elif is_init:
            if VALUES_ONLY_FROM_DB:  # todo: only if it is new
                raise AssertionError(f"VALUES_ONLY_FROM_DB: {name} is not in DB")
            set_new_value(name, ALL[name].default, version=ALL[name].version)

    if USER_DEFINED_TYPES:
        reset_user_values(db)


def check_update() -> None:
    """
    check if checksum in the cache backend is the same as the checksum in the local thread

    if not, the values from the database will be loaded
    """
    if not is_populated():
        return

    if TRIGGER.check():
        set_populated(False)


def recalc_checksums():
    """
    recalculate the checksums in the cache backend
    """
    TRIGGER.db_changed()



"""
context managers for the content settings, but not all `defaults` context manager can be found in `content_settings.defaults.context.defaults`
"""

from contextlib import ContextDecorator


class content_settings_context(ContextDecorator):
    """
    context manager that overwrites settings in the context.

    `**kwargs` for the context manager are the settings to overwrite, where key is a setting name and value is a raw value of the setting.

    outside of the content_settings module can be used for testing.

    `_raise_errors: bool = True` - if False, then ignore errors when applying value of the setting.
    """

    def __init__(self, **values) -> None:
        self.raise_errors = values.pop("_raise_errors", True)
        super().__init__()
        self.values_to_update = values
        self.prev_values = {}
        self.prev_types = {}

    def __enter__(self):
        from content_settings.caching import set_new_value, set_new_type

        for name, new_value in self.values_to_update.items():
            if isinstance(new_value, tuple):
                new_value, *type_define = new_value
                try:
                    self.prev_types[name] = set_new_type(name, *type_define)
                except:
                    if self.raise_errors:
                        raise
            try:
                self.prev_values[name] = set_new_value(name, new_value)
            except:
                if self.raise_errors:
                    raise

    def __exit__(self, *exc):
        from content_settings.caching import (
            set_new_value,
            replace_user_type,
            delete_user_value,
        )

        for name, new_value in self.prev_values.items():
            if name in self.prev_types:
                if self.prev_types[name] is None:
                    delete_user_value(name)
                    continue
                replace_user_type(name, self.prev_types[name])
            set_new_value(name, new_value)



"""
Django Models for the content settings.
"""

from collections import defaultdict

from django.db import models
from django.core.validators import RegexValidator
from django.utils import timezone
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.contrib.auth import get_user_model
from django.urls import reverse

from typing import Optional
from datetime import timedelta

User = get_user_model()


class ContentSetting(models.Model):
    """
    The main model for the content settings. Is stores all of the raw values for the content settings.
    """

    id = models.AutoField(primary_key=True, verbose_name=_("ID"))
    name = models.CharField(
        max_length=200,
        unique=True,
        validators=[
            RegexValidator(
                regex=r"^[A-Z][A-Z0-9_]*$",
                message=_(
                    "Value must be uppercase, can include numbers and underscores, and cannot start with a number."
                ),
            ),
        ],
        verbose_name=_("Name"),
    )
    value = models.TextField(blank=True, verbose_name=_("Value"))
    version = models.CharField(max_length=50, null=True, verbose_name=_("Version"))
    help = models.TextField(blank=True, null=True, verbose_name=_("Help"))
    tags = models.TextField(blank=True, null=True, verbose_name=_("Tags"))
    user_defined_type = models.CharField(
        max_length=50,
        null=True,
        default=None,
        db_index=True,
        verbose_name=_("User Defined Type"),
    )

    class Meta:
        ordering = ("name",)
        permissions = [
            ("can_preview_on_site", _("Can preview on site")),
        ]

    @property
    def tags_set(self):
        """
        tags field stores tags in a newline separated format. The property returns a set of tags.
        """
        if not self.tags:
            return set()

        return set(tag for tag in self.tags.splitlines() if tag.strip())

    def __str__(self):
        return self.name


class HistoryContentSetting(models.Model):
    """
    The model for the history of the content settings. Is used to store the history of changes for the content settings such as changed/added/removed.

    The the generation of the history is done in two steps.
    First step is to create a record when the setting is changed.
    Second step is to assign other changing parameters such as by_user.
    """

    CHANGED_CHOICES = (
        (True, _("Changed")),
        (False, _("Added")),
        (None, _("Removed")),
    )
    BY_USER_CHOICES = (
        (True, _("by user")),
        (False, _("by app")),
        (None, _("unknown")),
    )

    id = models.AutoField(primary_key=True, verbose_name=_("ID"))
    created_on = models.DateTimeField(
        null=False, default=timezone.now, verbose_name=_("Created On")
    )

    name = models.CharField(max_length=200, verbose_name=_("Name"))
    value = models.TextField(blank=True, null=True, verbose_name=_("Value"))
    version = models.CharField(max_length=50, null=True, verbose_name=_("Version"))

    was_changed = models.BooleanField(
        null=True, default=True, choices=CHANGED_CHOICES, verbose_name=_("Was Changed")
    )
    by_user = models.BooleanField(
        null=True, default=None, choices=BY_USER_CHOICES, verbose_name=_("By User")
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        verbose_name=_("User"),
    )
    help = models.TextField(blank=True, null=True, verbose_name=_("Help"))
    tags = models.TextField(blank=True, null=True, verbose_name=_("Tags"))
    user_defined_type = models.CharField(
        max_length=50, null=True, default=None, verbose_name=_("User Defined Type")
    )

    @property
    def user_defined_type_display(self):
        from .conf import USER_DEFINED_TYPES_NAME

        if not self.user_defined_type:
            return "---"

        if self.user_defined_type in USER_DEFINED_TYPES_NAME:
            return USER_DEFINED_TYPES_NAME[self.user_defined_type]

        return f"{self.user_defined_type} (unknown)"

    class Meta:
        ordering = ("-id",)
        indexes = [
            models.Index(fields=["name", "-id"], name="static_cont_name_d40020_idx"),
            models.Index(
                fields=["created_on", "user_id", "by_user"], name="static_batch_idx"
            ),
        ]

    @classmethod
    def update_last_record_for_name(cls, name: str, user: Optional[User] = None):
        """
        Update the last record with the information about the source of the update.
        """
        last_setting = cls.objects.filter(name=name).first()
        if not last_setting or last_setting.by_user is not None:
            return

        last_setting.user = user
        last_setting.by_user = user is not None
        last_setting.save()

    @property
    def admin_url_batch_changes(self):
        params = [
            f"created_on__gte={(self.created_on - timedelta(minutes=1)).replace(tzinfo=None).isoformat()}",
            f"created_on__lte={(self.created_on + timedelta(minutes=1)).replace(tzinfo=None).isoformat()}",
        ]

        if self.user:
            params.append(f"user_id={self.user_id}")

        if self.by_user is None:
            params.append("by_user__isnull=1")
        else:
            params.append(f"by_user={int(self.by_user)}")

        return (
            reverse("admin:content_settings_historycontentsetting_changelist")
            + "?"
            + "&".join(params)
        )

    @classmethod
    def gen_unique_records(cls, name):
        """
        The current issue is that sometimes the same setting is changed multiple times in a row.
        This method is used to generate unique records for the history.
        """
        next_obj = None
        records = cls.objects.filter(name=name)
        if not records.exists():
            return []

        for obj in records:
            if next_obj is None:
                next_obj = obj
                continue

            if (
                next_obj.value != obj.value
                or next_obj.was_changed != obj.was_changed
                or next_obj.tags != obj.tags
                or next_obj.help != obj.help
                or next_obj.user_defined_type != obj.user_defined_type
            ):
                yield next_obj

            next_obj = obj

        yield obj


class UserTagSetting(models.Model):
    """
    User can assign personal tags to the settings for extending tags-filtering functionality.
    The model contains those assignees.

    The allowed tags to assign in Django Admin panel can be found in `CONTENT_SETTINGS_USER_TAGS` django setting.
    """

    id = models.AutoField(primary_key=True, verbose_name=_("ID"))
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, verbose_name=_("User")
    )
    name = models.CharField(max_length=200, verbose_name=_("Name"))
    tag = models.CharField(max_length=200, verbose_name=_("Tag"))

    class Meta:
        unique_together = (("user", "name", "tag"),)

    def __str__(self):
        return f"{self.user} - {self.name} - {self.tag}"

    @classmethod
    def get_user_settings(cls, user):
        settings = defaultdict(set)
        for name, tag in cls.objects.filter(user=user).values_list("name", "tag"):
            settings[name].add(tag)
        return settings


class UserPreview(models.Model):
    """
    The user is allowed to preview settings before applying.

    The model contains the information of which settings are currently previewing.
    """

    id = models.AutoField(primary_key=True, verbose_name=_("ID"))
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, verbose_name=_("User")
    )
    name = models.CharField(max_length=200, verbose_name=_("Name"))
    from_value = models.TextField(verbose_name=_("From Value"), null=True)
    value = models.TextField(verbose_name=_("Value"))
    from_user_defined_type = models.CharField(
        max_length=50, null=True, verbose_name=_("From User Defined Type")
    )
    user_defined_type = models.CharField(
        max_length=50, null=True, verbose_name=_("User Defined Type")
    )
    from_tags = models.TextField(verbose_name=_("From Tags"), null=True)
    tags = models.TextField(verbose_name=_("Tags"), null=True)
    from_help = models.TextField(verbose_name=_("From Help"), null=True)
    help = models.TextField(verbose_name=_("Help"), null=True)

    @classmethod
    def get_context_dict(cls, user: User):
        return {
            v.name: (
                (v.value, v.user_defined_type, v.tags, v.help)
                if v.user_defined_type
                else v.value
            )
            for v in cls.objects.filter(user=user)
        }

    @classmethod
    def add_by_user(cls, user: User, name: str, **kwargs):
        """
        Adding the setting to the user's preview settings.
        """
        from_value = ContentSetting.objects.filter(name=name).first()
        try:
            preview_setting = cls.objects.get(user=user, name=name)
        except cls.DoesNotExist:
            preview_setting = cls.objects.create(
                user=user,
                name=name,
                **kwargs,
                **(
                    {"from_" + key: getattr(from_value, key) for key in kwargs.keys()}
                    if from_value
                    else {}
                ),
            )
        else:
            for key, value in kwargs.items():
                setattr(preview_setting, key, value)
            if from_value:
                for key in kwargs.keys():
                    setattr(preview_setting, "from_" + key, getattr(from_value, key))
            preview_setting.save()

        UserPreviewHistory.user_record(
            preview_setting, UserPreviewHistory.STATUS_CREATED
        )
        return preview_setting

    def apply(self):
        """
        Applying the preview setting into an actual setting.

        Works for:
        * userdefined settings
        * non-userdefined settings
        * userdefined preview for non-exist setting
        """
        setting = ContentSetting.objects.filter(name=self.name).first()
        is_userdefined = bool(self.user_defined_type)
        assert (
            setting is not None or is_userdefined
        ), "Setting should be set for non-userdefined settings"
        assert (is_userdefined and (not setting or setting.user_defined_type)) or (
            not is_userdefined and setting and not setting.user_defined_type
        ), "Setting should non-userdefined for userdefined preview"

        if is_userdefined:
            if not setting:
                setting = ContentSetting(
                    name=self.name,
                    user_defined_type=self.user_defined_type,
                    tags=self.tags,
                    help=self.help,
                )
            else:
                setting.tags = self.tags
                setting.help = self.help
                setting.user_defined_type = self.user_defined_type
        setting.value = self.value
        setting.save()
        HistoryContentSetting.update_last_record_for_name(setting.name, self.user)
        UserPreviewHistory.user_record(self, UserPreviewHistory.STATUS_APPLIED)

    class Meta:
        unique_together = (("user", "name"),)

    def __str__(self):
        return f"{self.user} - {self.name}"


class UserPreviewHistory(models.Model):
    """
    Contains history of the user's preview settings. Because the settings can also change logic, so we want to keep the history of the settings for future investigations.
    """

    STATUS_CREATED = 0
    STATUS_APPLIED = 10
    STATUS_REMOVED = 20
    STATUS_IGNORED = 30

    id = models.AutoField(primary_key=True, verbose_name=_("ID"))
    created_on = models.DateTimeField(null=False, default=timezone.now)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    name = models.CharField(max_length=200)
    value = models.TextField(null=True)
    user_defined_type = models.CharField(max_length=50, null=True)
    tags = models.TextField(null=True)
    help = models.TextField(null=True)

    status = models.IntegerField(
        default=STATUS_CREATED,
        choices=(
            (STATUS_CREATED, _("Created")),
            (STATUS_APPLIED, _("Applied")),
            (STATUS_REMOVED, _("Removed")),
            (STATUS_IGNORED, _("Ignored")),
        ),
    )

    @classmethod
    def user_record(cls, preview_setting: UserPreview, status: int = STATUS_CREATED):
        """
        Making a record in the history of the user's preview settings.
        """
        return cls.user_record_by_name(
            user=preview_setting.user,
            name=preview_setting.name,
            value=preview_setting.value,
            user_defined_type=preview_setting.user_defined_type,
            tags=preview_setting.tags,
            help=preview_setting.help,
            status=status,
        )

    @classmethod
    def user_record_by_name(
        cls,
        user: User,
        name: str,
        value: str,
        user_defined_type: Optional[str] = None,
        tags: Optional[str] = None,
        help: Optional[str] = None,
        status: int = STATUS_CREATED,
    ):
        """
        Making a record in the history of the user's preview settings by the name of the setting and the value.
        """
        return cls.objects.create(
            user=user,
            name=name,
            value=value,
            status=status,
            user_defined_type=user_defined_type,
            tags=tags,
            help=help,
        )

    def __str__(self):
        return f"{self.user} - {self.name}"



from enum import Enum, auto
from typing import Callable, Union


class PREVIEW(Enum):
    HTML = auto()
    TEXT = auto()
    PYTHON = auto()
    NONE = auto()


class required:
    pass


class optional:
    pass


def pre(value: str) -> str:
    return "<pre>{}</pre>".format(str(value).replace("<", "&lt;"))


class BaseSetting:
    pass


TCallableStr = Union[str, Callable]



"""
the functions that can be used for `CONTENT_SETTINGS_TAGS` and generate tags for the content settings based on setting name, type and value.
"""

from .settings import TAG_CHANGED
from .store import cs_has_app, get_app_name
from .types import BaseSetting
from typing import Set


def changed(name: str, cs_type: BaseSetting, value: str) -> Set[str]:
    """
    returns a tag `changed` if the value of the setting is different from the default value.

    the name of the tag can be changed in `CONTENT_SETTINGS_TAG_CHANGED` django setting.
    """
    return set() if value == cs_type.default else set([TAG_CHANGED])


def app_name(name: str, cs_type: BaseSetting, value: str) -> Set[str]:
    """
    returns a tag with the name of the app that uses the setting.
    """
    return set([get_app_name(name)]) if cs_has_app(name) else set()



"""
# Cache Trigger is a backend to sending a signal of when py-object(s) should be updated
"""

from typing import Any, Dict, Set, Optional, List
import hashlib
from functools import cached_property
from threading import local

from django.core.cache import caches

from . import __version__
from .settings import (
    USER_DEFINED_TYPES,
)


class ThreadLocalData(local):
    ALL_VALUES_CHECKSUM: str = ""


DATA = ThreadLocalData()


class BaseCacheTrigger:
    def __init__(self, params: dict) -> None:
        pass

    def check(self):
        """
        check if the trigger condition is met.
        """
        raise NotImplementedError

    def reset(self):
        """
        reset the trigger state.

        Called during the first run of the system.
        """
        raise NotImplementedError

    def get_form_checksum(self):
        """
        returns checksum that uses for checksum validation during form submit
        """
        raise NotImplementedError

    def db_changed(self):
        """
        recalculate the trigger state for the current DB state and save it to the outside source (cache backend).

        Called when the db value is changed.
        """
        raise NotImplementedError


class VersionChecksum(BaseCacheTrigger):
    """
    Arguments:
        * cache_backend (str, default="default"): cache backend name
        * cache_timeout (int, default=60 * 60 * 24): cache timeout
        * spliter (str, default="::"): spliter for the checksum
        * key_prefix (str, default="CS_CHECKSUM_"): key prefix for the checksum

    The algorithm:

        * during the first run, the checksum is calculated for all the values from the DB and saved to the cache backend
        * the cache key is calculated based on the current versions of all of the settings.
        * if the DB values are changed it saves the new checksum under the same cache key (so only instances with the same configuration will see the change)
        * if the checksum is changed comparing to the checksum in the local storage
    """

    def __init__(
        self,
        cache_backend: str = "default",
        cache_timeout: int = 60 * 60 * 24,
        spliter: str = "::",
        key_prefix: str = "CS_CHECKSUM_",
    ):
        self.cache_backend = caches[cache_backend]
        self.cache_timeout = cache_timeout
        self.spliter = spliter
        self.key_prefix = key_prefix + __version__
        self.last_checksum_from_cache = None

    def hash_value(self, value: str) -> str:
        """
        generate md5 hash for a string
        """
        return hashlib.md5(value.encode("utf-8")).hexdigest()

    def dict_checksum(self, values: Dict[str, Any]) -> str:
        """
        generate md5 hash for a dict with keys and values as strings
        """
        return self.hash_value(
            self.spliter.join(
                f"{name}{self.spliter}{values[name]}" for name in sorted(values.keys())
            )
        )

    @cached_property
    def cache_key(self) -> str:
        """
        returns cache key for checksum values (one should not be changed over time)
        """
        from .conf import ALL

        return (
            self.key_prefix
            + self.dict_checksum(
                {
                    name: ALL[name].version
                    for name in ALL.keys()
                    if not ALL[name].constant
                }
            )
            + self.user_cache_key_prefix
        )

    @property
    def user_cache_key_prefix(self) -> str:
        """
        returns cache key for user defined types (one should not be changed over time) if user defined types are not used, returns None
        """
        if not USER_DEFINED_TYPES:
            return ""

        from .conf import USER_DEFINED_TYPES_INITIAL

        return self.dict_checksum(
            {k: v.version for k, v in USER_DEFINED_TYPES_INITIAL.items()}
        )

    def set_local_checksum(self, value: Optional[str] = None) -> None:
        """
        calculate and set checksum in the local thread
        """
        if value is None:
            value = self.calc_checksum()
        DATA.ALL_VALUES_CHECKSUM = value

    def get_local_checksum(self) -> str:
        return DATA.ALL_VALUES_CHECKSUM

    def get_form_checksum(self):
        return self.get_local_checksum()

    def calc_checksum(self) -> str:
        from .models import ContentSetting

        return self.dict_checksum(
            {
                cs.name: cs.value + (cs.tags or "") + (cs.version or "")
                for cs in ContentSetting.objects.all()
            }
        )

    def push_checksum(self, value: Optional[str] = None) -> None:
        """
        save to cache backend the checksum
        """

        if value is None:
            value = DATA.ALL_VALUES_CHECKSUM

        self.cache_backend.set(self.cache_key, value, self.cache_timeout)

    def get_checksum_from_cache(self) -> Optional[str]:
        """
        get the checksum from the cache backend
        """
        return self.cache_backend.get(self.cache_key)

    def check(self):
        self.last_checksum_from_cache = self.get_checksum_from_cache()
        return (
            self.last_checksum_from_cache is not None
            and self.get_local_checksum() != self.last_checksum_from_cache != ""
        )

    def reset(self):
        if self.last_checksum_from_cache is None:
            self.set_local_checksum()
            self.push_checksum()
        else:
            self.set_local_checksum(self.last_checksum_from_cache)

    def db_changed(self):
        self.push_checksum(self.calc_checksum())



"""
Types that convert a string into a list of values.
"""

from typing import Optional, List, Callable, Generator, Iterable

from django.utils.translation import gettext as _

from .basic import (
    SimpleText,
    SimpleString,
    PREVIEW,
)
from .each import EachMixin, Item


def f_empty(value: str) -> Optional[str]:
    value = value.strip()
    if not value:
        return None
    return value


def f_comment(prefix: str) -> Callable[[str], Optional[str]]:
    def _(value: str) -> Optional[str]:
        if value.strip().startswith(prefix):
            return None
        return value

    return _


class SimpleStringsList(SimpleText):
    """
    Split a text into a list of strings.

    * comment_starts_with (default: #): if not None, the lines that start with this string are removed
    * filter_empty (default: True): if True, empty lines are removed
    * split_lines (default: \n): the string that separates the lines
    * filters (default: None): a list of additional filters to apply to the lines.
    """

    comment_starts_with: Optional[str] = "#"
    filter_empty: bool = True
    split_lines: str = "\n"
    filters: Optional[Iterable[Callable]] = None
    admin_preview_as: PREVIEW = PREVIEW.PYTHON

    def __init__(self, *args, **kwargs) -> None:
        from collections.abc import Iterable

        super().__init__(*args, **kwargs)
        assert isinstance(self.comment_starts_with, (str, type(None)))
        assert isinstance(self.filter_empty, bool)
        assert isinstance(self.split_lines, str)
        assert isinstance(self.filters, (Iterable, type(None)))

    def get_filters(self):
        """
        Get the filters based on the current configuration.

        * If filters is not None, it is returned.
        * If filter_empty is True, f_empty is added to the filters.
        * If comment_starts_with is not None, f_comment is added to the filters.
        """
        if self.filters is not None:
            filters = [*self.filters]
        else:
            filters = []

        if self.filter_empty:
            filters.append(f_empty)

        if self.comment_starts_with:
            filters.append(f_comment(self.comment_starts_with))

        return filters

    def get_help_format(self) -> Generator[str, None, None]:
        yield _("List of values with the following format:")
        yield "<ul>"
        if self.split_lines == "\n":
            yield "<li>"
            yield _("each line is a new value")
            yield "</li>"
        else:
            yield "<li>"
            yield _("%(split_lines)s separates values") % {
                "split_lines": self.split_lines
            }
            yield "</li>"
        yield "<li>"
        yield _("strip spaces from the beginning and from the end of the value")
        yield "</li>"
        yield "<li>"
        yield _("remove empty values")
        yield "</li>"
        if self.comment_starts_with:
            yield "<li>"
            yield _("use %(comment_starts_with)s to comment a line") % {
                "comment_starts_with": self.comment_starts_with
            }
            yield "</li>"
        if self.filter_empty:
            yield "<li>"
            yield _("empty values are removed")
            yield "</li>"
        yield "</ul>"

    def filter_line(self, line: str) -> str:
        for f in self.get_filters():
            line = f(line)
            if line is None:
                break
        return line

    def gen_to_python(self, value: str) -> Generator[str, None, None]:
        """
        Converts a string value into a generator of filtered lines.
        """
        lines = value.split(self.split_lines)
        for line in lines:
            line = self.filter_line(line)

            if line is not None:
                yield line

    def to_python(self, value: str) -> List[str]:
        return list(self.gen_to_python(value))


class TypedStringsList(EachMixin, SimpleStringsList):
    line_type = SimpleString()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.each = Item(self.line_type)



"""
the module collects all settings from all apps and makes them available as `content_settings` object.

It has the following global variables:

* `USER_DEFINED_TYPES_INSTANCE: Dict[str, Callable]` - key is the slug of the user defined setting, value is a function that returns a type with tags and help text
* `USER_DEFINED_TYPES_INITIAL: Dict[str, BaseSetting]` - key is a slug of the user defined type, value is a type without tags and help text
* `USER_DEFINED_TYPES_NAME: Dict[str, str]` - the name of the user defined types
* `ALL: Dict[str, BaseSetting]` - the all registereg settings types
* `CALL_TAGS: Optional[List[Callable]]` - the list of function that is taken from `CONTENT_SETTINGS_TAGS` setting and used to generate tags for settings.
* `CONSTANTS: Set[str]` - a set of names of content settings that are constants. Those are not stored in DB.
"""

from importlib import import_module
from functools import partial
from typing import Any, Callable, Optional, List, Set, Dict, Tuple, Union
from functools import cached_property

from django.apps import apps
from django.core.exceptions import ValidationError

from .types.basic import BaseSetting
from .caching import (
    get_value,
    get_py_value,
    get_type_by_name,
    get_all_names,
    get_form_checksum,
)
from .settings import USER_DEFINED_TYPES, TAGS, CHAIN_VALIDATE
from .store import (
    add_app_name,
    get_admin_head,
    add_admin_head,
    get_admin_raw_js,
    register_prefix,
    PREFIXSES,
)
from .context_managers import content_settings_context
from .utils import import_object

USER_DEFINED_TYPES_INSTANCE: Dict[str, Callable] = {}
USER_DEFINED_TYPES_INITIAL: Dict[str, BaseSetting] = {}
USER_DEFINED_TYPES_NAME: Dict[str, str] = {}
ALL: Dict[str, BaseSetting] = {}
CALL_TAGS: Optional[List[Callable]] = None
CONSTANTS: Set[str] = set()


def is_constant(name: str) -> bool:
    """
    check if the setting is constant
    """
    return name in CONSTANTS


def get_call_tags() -> List[Callable]:
    """
    returns list of functions from `CONTENT_SETTINGS_TAGS` setting that are used to generate tags for settings.
    the result is cached in `CALL_TAGS` variable.
    """
    global CALL_TAGS

    if CALL_TAGS is not None:
        return CALL_TAGS

    CALL_TAGS = []
    for func_tag in TAGS:
        if isinstance(func_tag, str):
            func_tag = import_object(func_tag)
        elif callable(func_tag):
            pass
        else:
            raise AssertionError(
                f"CONTENT_SETTINGS_TAGS: {func_tag} should be str or callable"
            )
        CALL_TAGS.append(func_tag)
    return CALL_TAGS


def gen_tags(name: str, cs_type: BaseSetting, value: Any) -> Set[str]:
    """
    generate tags based on `CONTENT_SETTINGS_TAGS` setting.
    """
    tags = set()
    for func_tag in get_call_tags():
        tags |= func_tag(name, cs_type, value)
    return tags


@register_prefix("lazy")
def lazy_prefix(name: str, suffix: str) -> Any:
    """
    lazy__ prefix that gives a lazy proxy object by the name of the setting.
    """
    return get_type_by_name(name).lazy_give(lambda: get_value(name, suffix), suffix)


@register_prefix("type")
def type_prefix(name: str, suffix: str) -> Any:
    """
    type__ prefix that return setting type by the name of the setting.
    """
    assert not suffix, "type prefix can not have suffix"

    return get_type_by_name(name)


@register_prefix("startswith")
def startswith_prefix(name: str, suffix: str) -> Dict[str, Any]:
    """
    startswith__ prefix that returns all settings as a dict (setting name: setting value) that start with the given name.
    """
    return {
        k: get_value(k, suffix) for k in dir(content_settings) if k.startswith(name)
    }


@register_prefix("withtag")
def withtag_prefix(name: str, suffix: str) -> Dict[str, Any]:
    """
    withtag__ prefix that returns all settings as a dict (setting name: setting value) that have the given tag.
    """
    return {
        k: get_value(k, suffix)
        for k in dir(content_settings)
        if name in get_type_by_name(k).get_tags()
        or name.lower() in get_type_by_name(k).get_tags()
    }


def split_attr(value: str) -> Tuple[Optional[str], str, Optional[str]]:
    """
    splits the name of the attr on 3 parts: prefix, name, suffix

    * prefix should be registered by register_prefix
    * name should be uppercase
    * suffix can be any string, but not uppercase
    """
    prefix = None
    parts = value.split("__")

    if parts[0] in PREFIXSES:
        prefix = parts.pop(0)

    assert len(parts), f"Invalid attribute name: {value}; can not be only prefix"

    name = parts.pop(0)
    assert name.isupper(), f"Invalid attribute name: {value}; name should be uppercase"

    while parts:
        if not parts[0].isupper():
            break

        name += "__" + parts.pop(0)

    if len(parts):
        return prefix, name, "__".join(parts).lower()

    return prefix, name, None


def validate_all_with_context(context: Dict[str, Any]):
    """
    validate all settings with the given context to make sure all of them are valid.

    Do not perform if `CONTENT_SETTINGS_CHAIN_VALIDATE = False`
    """
    if not CHAIN_VALIDATE:
        return

    with content_settings_context(**context):
        for name, cs_type in ALL.items():
            try:
                if name in context:
                    cs_type.validate_raw_value(context[name])
                else:
                    cs_type.validate(get_py_value(name))
            except Exception as e:
                raise ValidationError(f"Error validating {name}: {e}")


def get_str_tags(
    cs_name: str, cs_type: BaseSetting, value: Optional[str] = None
) -> str:
    """
    get tags as a text (joined by `\n`) for specific setting type. name and value are used to generate content tags.

    from saving in DB.
    """
    tags = cs_type.get_tags()
    if not cs_type.user_defined_slug:
        tags |= cs_type.get_content_tags(
            cs_name, cs_type.default if value is None else value
        )
    return "\n".join(sorted(tags))


def set_initial_values_for_db(apply: bool = False) -> List[Tuple[str, str]]:
    """
    sync settings with DB.
        * creates settings that are not in DB
        * updates settings that are in DB but have different attributes such as help text or tags
        * deletes settings that are in DB but are not in ALL

    attribute `apply` is used to apply changes in DB immediately. Can be used in tests.
    """
    from content_settings.models import ContentSetting, HistoryContentSetting

    changes = []

    def execute(name, key, func):
        changes.append((name, key))
        if apply:
            func()
            HistoryContentSetting.update_last_record_for_name(name)

    def execute_update_obj(name, cs, show="update", **kwargs):
        def _up():
            for k, v in kwargs.items():
                setattr(cs, k, v)
            cs.save()

        execute(name, show, _up)

    for k, cs_type in ALL.items():
        if cs_type.constant:
            continue

        try:
            ContentSetting.objects.get(name=k)
        except ContentSetting.DoesNotExist:
            execute(
                k,
                "create",
                lambda: ContentSetting.objects.create(
                    name=k,
                    value=cs_type.default,
                    version=cs_type.version,
                    tags=get_str_tags(k, cs_type),
                    help=cs_type.get_help(),
                ),
            )

    for cs in ContentSetting.objects.all():
        if cs.name in ALL:
            cs_type = ALL[cs.name]
            if cs_type.constant:
                execute(cs.name, "delete", lambda: cs.delete())
                continue

            assert (
                not cs.user_defined_type or cs_type.overwrite_user_defined
            ), f"{cs.name} is not a code setting and not overwrite_user_defined"

            if cs.version != cs_type.version:
                execute_update_obj(
                    cs.name,
                    cs,
                    value=cs_type.default,
                    version=cs_type.version,
                    user_defined_type=None,
                )

            if cs.user_defined_type:
                execute_update_obj(
                    cs.name,
                    cs,
                    user_defined_type=None,
                    show="adjust",
                )

            str_tags = get_str_tags(cs.name, cs_type, cs.value)
            str_help = cs_type.get_help()

            if cs.tags != str_tags or cs.help != str_help:
                execute_update_obj(
                    cs.name,
                    cs,
                    tags=str_tags,
                    help=str_help,
                    show="adjust",
                )

        else:
            if cs.user_defined_type:
                if cs.user_defined_type not in USER_DEFINED_TYPES_INSTANCE:
                    execute(cs.name, "delete", lambda: cs.delete())
                elif (
                    cs.version
                    != USER_DEFINED_TYPES_INITIAL[cs.user_defined_type].version
                ):
                    cs_type = USER_DEFINED_TYPES_INSTANCE[cs.user_defined_type]()
                    execute_update_obj(
                        cs.name, cs, value=cs_type.default, version=cs_type.version
                    )
            else:
                execute(cs.name, "delete", lambda: cs.delete())

    return changes


#################################################
# Loop for defining ALL types and other globals #
#################################################

if USER_DEFINED_TYPES:
    for slug, imp_line, name in USER_DEFINED_TYPES:
        type_class = import_object(imp_line)
        USER_DEFINED_TYPES_INSTANCE[slug] = partial(
            type_class, user_defined_slug=slug, version=type_class.version
        )
        initial_val = USER_DEFINED_TYPES_INSTANCE[slug]()
        USER_DEFINED_TYPES_INITIAL[slug] = initial_val
        add_admin_head(initial_val)
        USER_DEFINED_TYPES_NAME[slug] = name


for app_config in apps.app_configs.values():
    app = app_config.name
    try:
        content_settings = import_module(app + ".content_settings")
    except ImportError as e:
        if e.name != app + ".content_settings":
            raise
        continue
    for attr in dir(content_settings):
        if not attr.isupper():
            continue

        val = getattr(content_settings, attr)

        if not isinstance(val, BaseSetting):
            continue

        assert attr not in ALL, "Content Setting {} defined twice".format(attr)

        if attr in ALL and not ALL[attr].user_defined_slug:
            raise AssertionError("Overwriting content setting {}".format(attr))

        assert (
            not val.user_defined_slug
        ), "Do not set user_defined_slug in content_settings.py"

        if val.constant:
            CONSTANTS.add(attr)

        ALL[attr] = val
        add_admin_head(val)
        add_app_name(attr, app)


##############################################
# The Main Content Settings Class and object #
##############################################


class _Settings:
    """
    the main object that uses for getting settings for cache.
    """

    def __getattr__(self, value):
        prefix, name, suffix = split_attr(value)
        if prefix:
            assert (
                prefix in PREFIXSES
            ), f"Invalid attribute name: {value}; prefix not found"
            return PREFIXSES[prefix](name, suffix)
        return get_value(name, suffix)

    def __setattr__(self, name: str, value: Union[str, Tuple]) -> None:
        prefix, name, suffix = split_attr(name)
        assert not prefix, "Can not set attribute with prefix"
        assert not suffix, "Can not set attribute with suffix"
        assert isinstance(value, str) or (
            isinstance(value, tuple) and USER_DEFINED_TYPES
        ), "You can set only raw value or user defined type"

        from .caching import set_new_db_value

        if isinstance(value, str):
            set_new_db_value(name, value)
        else:
            assert USER_DEFINED_TYPES, "User defined types are not enabled"
            set_new_db_value(name, *value)

    def __dir__(self):
        """
        dir() returns all settings names
        """
        return get_all_names()

    def __contains__(self, value):
        _, name, suffix = split_attr(value)
        cs_type = get_type_by_name(name)
        return cs_type is not None and cs_type.can_suffix(suffix)

    @property
    def form_checksum(self):
        """
        the current checksum of the settings.

        used for validation of settings weren't changed over time.
        """
        return get_form_checksum()

    @cached_property
    def admin_head(self):
        """
        the admin head.
        """
        return get_admin_head()

    @cached_property
    def admin_raw_js(self):
        """
        the admin raw js.
        """
        return get_admin_raw_js()


content_settings = _Settings()



from contextlib import contextmanager
from typing import Iterator, Set, Dict, Any

from content_settings.settings import DEFAULTS as DEFAULTS_SETTINGS
from content_settings.types import BaseSetting
from content_settings.defaults.modifiers import TModifier

from .filters import any_name
from .modifiers import set_if_missing, add_tags, help_prefix, help_suffix, NotSet


DEFAULTS = [*DEFAULTS_SETTINGS]


@contextmanager
def defaults(*args, **kwargs):
    """
    Context manager for setting defaults.
    """
    DEFAULTS.append((any_name, *args, set_if_missing(**kwargs)))
    try:
        yield
    finally:
        DEFAULTS.pop()


@contextmanager
def default_tags(tags: Set[str]):
    """
    defaults context for setting default tags.
    """
    with defaults(add_tags(tags)):
        yield


@contextmanager
def default_help_prefix(prefix: str):
    """
    defaults context for setting default help prefix.
    """
    with defaults(help_prefix(prefix)):
        yield


@contextmanager
def default_help_suffix(suffix: str):
    """
    defaults context for setting default help suffix.
    """
    with defaults(help_suffix(suffix)):
        yield


def defaults_modifiers(setting: BaseSetting) -> Iterator[TModifier]:
    """
    Generator for all modifiers for the given setting.
    """
    for modifier in DEFAULTS:
        if not modifier[0](setting.__class__):
            continue
        yield from modifier[1:]


def update_defaults(setting: BaseSetting, kwargs: Dict[str, Any]):
    """
    Update paramas of the setting type by applying all of the modifiers from the defaults context.
    """
    type_kwargs = {
        name: getattr(setting, name)
        for name in dir(setting)
        if setting.can_assign(name)
    }
    updates = {}
    for modifier in defaults_modifiers(setting):
        updates.update(modifier(type_kwargs, updates, kwargs))

    if not updates:
        return kwargs

    return {
        **kwargs,
        **{
            k: v
            for k, v in updates.items()
            if setting.can_assign(k) and v is not NotSet and type_kwargs[k] != v
        },
    }



"""
defaults collections for using in `CONTENT_SETTINGS_DEFAULTS`.

For example:

```python
CONTENT_SETTINGS_DEFAULTS = [
    codemirror_python(),
    codemirror_json(),
]
```

Or:

```python
CONTENT_SETTINGS_DEFAULTS = [
    *codemirror_all(),
]
```
"""

from content_settings.functools import or_

from .modifiers import add_admin_head, add_widget_class
from .filters import full_name_exact

DEFAULT_CODEMIRROR_PATH = "https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.65.16/"


def _codemirror_raw_js(class_attr: str, mode: str):
    return f"""
    Array.from(document.getElementsByClassName("{class_attr}")).forEach((item, i) => {{
        const cm = CodeMirror.fromTextArea(item, {{
            lineNumbers: true,
            mode: "{mode}"
        }});
        cm.on("change", function(editor) {{
            item.value = editor.getValue();
        }});
        cm.on("focus", function(editor) {{
            item.dispatchEvent(new Event('focus'));
        }});
        cm.display.wrapper.after(item);
    }});
    """


def codemirror_python(
    path: str = DEFAULT_CODEMIRROR_PATH, class_attr: str = "codemirror_python"
):
    """
    Replace Textarea with CodeMirror for python code for SimpleEval and SimpleExec.
    """
    return (
        or_(
            full_name_exact("content_settings.types.template.SimpleEval"),
            full_name_exact("content_settings.types.template.SimpleExec"),
            full_name_exact("content_settings.types.template.SimpleExecNoCompile"),
        ),
        add_admin_head(
            css=[f"{path}codemirror.min.css"],
            js=[
                f"{path}codemirror.min.js",
                f"{path}mode/python/python.min.js",
            ],
            js_raw=[_codemirror_raw_js(class_attr, "python")],
        ),
        add_widget_class(class_attr),
    )


def codemirror_json(
    path: str = DEFAULT_CODEMIRROR_PATH, class_attr: str = "codemirror_json"
):
    """
    Replace Textarea with CodeMirror for json code for SimpleJSON.
    """
    return (
        full_name_exact("content_settings.types.markup.SimpleJSON"),
        add_admin_head(
            css=[f"{path}codemirror.min.css"],
            js=[
                f"{path}codemirror.min.js",
                f"{path}mode/javascript/javascript.min.js",
            ],
            js_raw=[_codemirror_raw_js(class_attr, "javascript")],
        ),
        add_widget_class(class_attr),
    )


def codemirror_yaml(
    path: str = DEFAULT_CODEMIRROR_PATH, class_attr: str = "codemirror_yaml"
):
    """
    Replace Textarea with CodeMirror for yaml code for SimpleYAML.
    """
    return (
        full_name_exact("content_settings.types.markup.SimpleYAML"),
        add_admin_head(
            css=[f"{path}codemirror.min.css"],
            js=[
                f"{path}codemirror.min.js",
                f"{path}mode/yaml/yaml.min.js",
            ],
            js_raw=[_codemirror_raw_js(class_attr, "yaml")],
        ),
        add_widget_class(class_attr),
    )


def codemirror_all(
    path: str = DEFAULT_CODEMIRROR_PATH, class_attr_prefix: str = "codemirror_"
):
    """
    Replace Textarea with CodeMirror for python, json and yaml code.
    """
    return [
        codemirror_python(path, f"{class_attr_prefix}python"),
        codemirror_json(path, f"{class_attr_prefix}json"),
        codemirror_yaml(path, f"{class_attr_prefix}yaml"),
    ]



"""
Module for exporting, previewing and importing content settings from and to JSON
"""

import json

from django.utils.translation import gettext as _
from django.contrib.auth import get_user_model

from typing import Iterable, Tuple, List, Optional, Dict

from .types import BaseSetting
from .models import ContentSetting, HistoryContentSetting, UserPreview
from .conf import (
    USER_DEFINED_TYPES_INITIAL,
    get_type_by_name,
    validate_all_with_context,
)
from .migrate import import_settings


User = get_user_model()


class Error(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__()

    def __str__(self) -> str:
        return self.message


def export_to_format(content_settings: Iterable[ContentSetting]) -> str:
    """
    Export content settings to JSON format
    """
    data = {}
    data_settings = data["settings"] = {}
    for cs in content_settings:

        if cs.user_defined_type:
            data_settings[cs.name] = {
                "value": cs.value,
                "tags": cs.tags,
                "help": cs.help,
                "version": cs.version,
                "user_defined_type": cs.user_defined_type,
            }
        else:
            data_settings[cs.name] = {
                "value": cs.value,
                "version": cs.version,
            }

    return json.dumps(data, indent=2)


def preview_data(data: dict, user: Optional[User] = None) -> Tuple[List, List, List]:
    """
    Validate data and return three lists: errors, applied, skipped

    Those list are used for previewing import and applying import.
    """
    errors = []
    applied = []
    skipped = []
    for name, value in data["settings"].items():
        try:
            if applied_row := applied_preview(name, value, user):
                applied.append(applied_row)
            else:
                skipped.append({"name": name, "reason": _("Value is the same")})
        except Exception as e:
            errors.append({"name": name, "reason": str(e)})
    return errors, applied, skipped


def applied_preview(
    name: str, value: dict, user: Optional[User] = None
) -> Optional[dict]:
    """
    the function returns applied element for previewing import.

    if function returns None, the setting is not applied, added to skipped list instead.

    if function raises an exception, the setting is not applied, added to errors list instead.
    """

    if "user_defined_type" in value:
        return applied_preview_user_defined_type(name, value, user)

    try:
        db_setting = ContentSetting.objects.get(name=name)
    except ContentSetting.DoesNotExist:
        raise Error(_("Setting does not exist"))

    cs_type = get_type_by_name(name)

    if cs_type is None:
        raise Error(_("Setting does not exist (type not found)"))

    for key in ["value", "version"]:
        if key not in value or not isinstance(value[key], str):
            raise Error(_('"%s" is not set or not a string' % key))

    if db_setting.value == value["value"]:
        return None

    if db_setting.version != value["version"]:
        raise Error(_("Version is not the same"))

    if user is not None and not cs_type.can_update(user):
        raise Error(_("You don't have permissions to update the setting"))

    cs_type.validate_value(value["value"])

    return {
        "name": name,
        "old_value": db_setting.value,
        "new_value": value["value"],
        "full": {"value": value["value"]},
    }


def applied_preview_user_defined_type(
    name: str, value: dict, user: Optional[User] = None
) -> Optional[dict]:
    """
    `applied_preview` for user defined type.
    """
    for key in ["value", "version", "tags", "help"]:
        if key not in value or not isinstance(value[key], str):
            raise Error('"%s" is not set or not a string' % key)

    cs_type = USER_DEFINED_TYPES_INITIAL.get(value["user_defined_type"])

    if cs_type is None:
        raise Error(_("Invalid user_defined_type"))

    if cs_type.version != value["version"]:
        raise Error(_("Invalid version"))

    if user is not None and not cs_type.can_update(user):
        raise Error(_("You don't have permissions to update the setting"))

    cs_type.validate_value(value["value"])

    try:
        db_setting = ContentSetting.objects.get(name=name)
    except ContentSetting.DoesNotExist:
        return {"name": name, "new_value": value, "full": value}

    if not db_setting.user_defined_type:
        raise Error(_("Setting is not user defined"))

    applied_new_value = {}
    applied_old_value = {}
    for key, in_value in value.items():
        if (
            key == "tags"
            and set(getattr(db_setting, key).splitlines()) != set(in_value.splitlines())
            or key != "tags"
            and getattr(db_setting, key) != in_value
        ):
            applied_new_value[key] = in_value
            applied_old_value[key] = getattr(db_setting, key)

    if not applied_new_value:
        return None

    return {
        "name": name,
        "old_value": applied_old_value,
        "new_value": applied_new_value,
        "full": value,
    }


def import_to(
    data: Dict, applied: List[Dict], preview: bool, user: Optional[User] = None
):
    """
    Import content settings from JSON data and previewed apply data. Arguments:
    - data: JSON data
    - applied: List of applied settings. The list is returned by import_preview_data.
    - user: User object
    - preview: If True, import to preview

    It raises Error if something is wrong.
    """
    prepared_names = {}
    for v in applied:
        data_setting = data["settings"][v["name"]]
        if "user_defined_type" in data_setting:
            prepared_names[v["name"]] = (
                data_setting["value"],
                data_setting["user_defined_type"],
                (
                    set(data_setting["tags"].splitlines())
                    if data_setting["tags"]
                    else set()
                ),
                data_setting["help"],
            )
        else:
            prepared_names[v["name"]] = v["new_value"]

    validate_all_with_context(prepared_names)

    if preview:
        assert user is not None, "User is required"
        for value in applied:
            UserPreview.add_by_user(
                user=user,
                name=value["name"],
                **{k: v for k, v in value["full"].items() if k != "version"},
            )
    else:
        import_settings(
            {"settings": {value["name"]: value["full"] for value in applied}},
            model_cs=ContentSetting,
            model_cs_history=HistoryContentSetting,
            user=user,
        )



"""
the module is used for collecting information.

the `APP_NAME_STORE` a dict `setting_name: app_name` is used to store the name of the app that uses the setting. Which later on can be used in `tags.app_name` to generate a tag with the name of the app.
"""

from typing import Callable, Dict, List

from .types import BaseSetting

APP_NAME_STORE: Dict[str, str] = {}  # setting_name: app_name

ADMIN_HEAD_CSS: List[str] = []
ADMIN_HEAD_JS: List[str] = []
ADMIN_HEAD_CSS_RAW: List[str] = []
ADMIN_HEAD_JS_RAW: List[str] = []

PREFIXSES: Dict[str, Callable] = {}


def add_app_name(cs_name: str, app_name: str) -> None:
    """
    add the name of the app that uses the setting.
    """
    APP_NAME_STORE[cs_name] = app_name


def cs_has_app(cs_name: str) -> bool:
    """
    check if the setting has an app name.
    """
    return cs_name in APP_NAME_STORE


def get_app_name(cs_name: str) -> str:
    """
    get the name of the app that uses the setting.
    """
    return APP_NAME_STORE[cs_name]


def add_admin_head_css(css_url: str) -> None:
    """
    add a css url to the admin head.
    """
    if css_url not in ADMIN_HEAD_CSS:
        ADMIN_HEAD_CSS.append(css_url)


def add_admin_head_js(js_url: str) -> None:
    """
    add a js url to the admin head.
    """
    if js_url not in ADMIN_HEAD_JS:
        ADMIN_HEAD_JS.append(js_url)


def add_admin_head_css_raw(css: str) -> None:
    """
    add a css code to the admin head.
    """
    if css not in ADMIN_HEAD_CSS_RAW:
        ADMIN_HEAD_CSS_RAW.append(css)


def add_admin_head_js_raw(js: str) -> None:
    """
    add a js code to the admin head.
    """
    if js not in ADMIN_HEAD_JS_RAW:
        ADMIN_HEAD_JS_RAW.append(js)


def add_admin_head(setting: BaseSetting) -> None:
    """
    add a setting to the admin head.
    """
    for css in setting.get_admin_head_css():
        add_admin_head_css(css)
    for js in setting.get_admin_head_js():
        add_admin_head_js(js)
    for css in setting.get_admin_head_css_raw():
        add_admin_head_css_raw(css)
    for js in setting.get_admin_head_js_raw():
        add_admin_head_js_raw(js)


def get_admin_head() -> str:
    """
    get the admin head.
    """

    def gen():
        for link in ADMIN_HEAD_CSS:
            yield f"<link rel='stylesheet' href='{link}' />"
        for link in ADMIN_HEAD_JS:
            yield f"<script src='{link}'></script>"
        for code in ADMIN_HEAD_CSS_RAW:
            yield f"<style>{code}</style>"

    return "".join(gen())


def get_admin_raw_js() -> str:
    """
    get the admin raw js.
    """

    def gen():
        for code in ADMIN_HEAD_JS_RAW:
            yield f"(function($){{{code}}})(django.jQuery);"

    return "".join(gen())


def register_prefix(name: str) -> Callable:
    """
    decorator for registration a new prefix
    """

    def _cover(func: Callable) -> Callable:
        assert name not in PREFIXSES
        PREFIXSES[name] = func
        return func

    return _cover



"""
EachMixin is the main mixin of the module, which allows types to have subtypes, that check, preview and converts the structure of the value.

For example `array.TypedStringsList`
"""

from enum import Enum, auto
from typing import Union, Any, Dict, Optional

from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _

from . import required, optional, PREVIEW, pre, BaseSetting


class BaseEach:
    def is_each(self, value: Any):
        return True

    def validate(self, value: Any):
        if not self.is_each(value):
            return
        self.each_validate(value)

    def to_python(self, value: Any):
        if not self.is_each(value):
            return value
        return self.each_to_python(value)

    def give(self, value: Any, suffix=None):
        if not self.is_each(value):
            return value
        return self.each_give(value, suffix)

    def give_python_to_admin(self, value: Any, name, **kwargs):
        if not self.is_each(value):
            return value
        return self.each_give_python_to_admin(value, name, **kwargs)

    def get_admin_preview_object(self, value, *args, **kwargs):
        if not self.is_each(value):
            return pre(value)
        return self.each_get_admin_preview_object(value, *args, **kwargs)

    def get_all_cs_types(self):
        return ()

    def get_admin_head_css(self):
        return sum(
            (cs.get_admin_head_css() for cs in self.get_all_cs_types()), start=()
        )

    def get_admin_head_js(self):
        return sum((cs.get_admin_head_js() for cs in self.get_all_cs_types()), start=())

    def get_admin_head_css_raw(self):
        return sum(
            (cs.get_admin_head_css_raw() for cs in self.get_all_cs_types()), start=()
        )

    def get_admin_head_js_raw(self):
        return sum(
            (cs.get_admin_head_js_raw() for cs in self.get_all_cs_types()), start=()
        )

    def can_suffix(self, suffix: Optional[str], value: Any) -> bool:
        return all(cs.can_suffix(suffix, value) for cs in self.get_all_cs_types())


TCSType = Union[BaseEach, BaseSetting]


class Item(BaseEach):
    """
    Converts each element of the array into a specific type `cs_type`
    """

    def __init__(self, cs_type: TCSType):
        self.cs_type = cs_type

    def is_each(self, value: Any):
        return isinstance(value, (list, tuple))

    def get_all_cs_types(self):
        return (self.cs_type,)

    def each_validate(self, value):
        for i, v in enumerate(value, start=1):
            try:
                self.cs_type.validate(v)
            except ValidationError as e:
                raise ValidationError(
                    _("item #%(i)s: %(e)s") % {"i": i, "e": e.message}
                )

    def each_to_python(self, value):
        ret = []
        for i, v in enumerate(value, start=1):
            try:
                ret.append(self.cs_type.to_python(v))
            except ValidationError as e:
                raise ValidationError(
                    _("item #%(i)s: %(e)s") % {"i": i, "e": e.message}
                )
        return ret

    def each_give_python_to_admin(self, value, name, **kwargs):
        ret = []
        for i, v in enumerate(value, start=1):
            try:
                ret.append(self.cs_type.give_python_to_admin(v, name, **kwargs))
            except ValidationError as e:
                raise ValidationError(
                    _("item #%(i)s: %(e)s") % {"i": i, "e": e.message}
                )
        return ret

    def get_help_format(self):
        yield _("A list of items. Each item should be: ")
        yield "<div class='subitem'>"
        yield from self.cs_type.get_help_format()
        yield "</div>"

    def each_give(self, value, suffix=None):
        return [self.cs_type.give(v, suffix) for v in value]

    def each_get_admin_preview_object(self, value, *args, **kwargs):
        return (
            "["
            + (
                ",".join(
                    [
                        "<div class='subitem'>"
                        + self.cs_type.get_admin_preview_object(v, *args, **kwargs)
                        + "</div>"
                        for v in value
                    ]
                )
            )
            + "]"
        )


class Keys(BaseEach):
    """
    Converts values of the specific keys into specific types `cs_types`
    """

    def __init__(self, **kwargs: Dict[str, TCSType]):
        self.cs_types = kwargs

    def get_all_cs_types(self):
        return tuple(self.cs_types.values())

    def is_each(self, value):
        return isinstance(value, dict)

    def each_validate(self, value):
        for k, v in value.items():
            if k not in self.cs_types:
                continue
            try:
                self.cs_types[k].validate(v)
            except ValidationError as e:
                raise ValidationError(_("key %(k)s: %(e)s") % {"k": k, "e": e.message})

        for k, v in self.cs_types.items():
            if k not in value:
                if v.default == required:
                    raise ValidationError(_("Missing required key %(k)s") % {"k": k})

    def each_to_python(self, value):
        ret = {}
        for k, v in value.items():
            if k in self.cs_types:
                try:
                    ret[k] = self.cs_types[k].to_python(v)
                except ValidationError as e:
                    raise ValidationError(
                        _("key %(k)s: %(e)s") % {"k": k, "e": e.message}
                    )
            else:
                ret[k] = v

        for k, v in self.cs_types.items():
            if k not in ret and v.default not in (optional, required):
                ret[k] = v.to_python(v.default)

        return ret

    def each_give_python_to_admin(self, value, name, **kwargs):
        ret = {}
        for k, v in value.items():
            if k in self.cs_types:
                try:
                    ret[k] = self.cs_types[k].give_python_to_admin(v, name, **kwargs)
                except ValidationError as e:
                    raise ValidationError(
                        _("key %(k)s: %(e)s") % {"k": k, "e": e.message}
                    )
            else:
                ret[k] = v

        for k, v in self.cs_types.items():
            if k not in ret and v.default not in (optional, required):
                ret[k] = v.give_python_to_admin(v.default, name, **kwargs)

        return ret

    def get_help_format(self):
        yield _("A dictionary. Keys: ")
        for k, v in self.cs_types.items():
            yield "<div class='subitem'>"
            yield f"<i>{k}</i> - "
            if v.default == required:
                yield _("(required) ")
            elif v.default == optional:
                yield _("(optional) ")
            else:
                yield _("(default: %(default)s) ") % {
                    "default": v.default if v.default else ("<i>" + _("empty") + "</i>")
                }
            yield from v.get_help_format()
            yield "</div>"

    def each_give(self, value, suffix=None):
        return {
            k: self.cs_types[k].give(v, suffix) if k in self.cs_types else v
            for k, v in value.items()
        }

    def each_get_admin_preview_object(self, value, *args, **kwargs):
        return (
            "{"
            + ",".join(
                [
                    f"<div class='subitem'><i>{k}</i>: {self.cs_types[k].get_admin_preview_object(v, *args, **kwargs) if k in self.cs_types else pre(v)}</div>"
                    for k, v in value.items()
                ]
            )
            + "}"
        )


class Values(BaseEach):
    """
    Converts each value of the given dict into `cs_type`
    """

    def __init__(self, cs_type: TCSType):
        self.cs_type = cs_type

    def get_all_cs_types(self):
        return (self.cs_type,)

    def is_each(self, value):
        return isinstance(value, dict)

    def each_validate(self, value):
        for k, v in value.items():
            try:
                self.cs_type.validate(v)
            except ValidationError as e:
                raise ValidationError(_("key %(k)s: %(e)s") % {"k": k, "e": e.message})

    def each_to_python(self, value):
        ret = {}

        for k, v in value.items():
            try:
                ret[k] = self.cs_type.to_python(v)
            except ValidationError as e:
                raise ValidationError(_("key %(k)s: %(e)s") % {"k": k, "e": e.message})

        return ret

    def each_give_python_to_admin(self, value, name, **kwargs):
        ret = {}

        for k, v in value.items():
            try:
                ret[k] = self.cs_type.give_python_to_admin(v, name, **kwargs)
            except ValidationError as e:
                raise ValidationError(_("key %(k)s: %(e)s") % {"k": k, "e": e.message})

        return ret

    def get_help_format(self):
        yield _("A dictionary. Each value should be: ")
        yield "<div class='subitem'>"
        yield from self.cs_type.get_help_format()
        yield "</div>"

    def each_give(self, value, suffix=None):
        return {k: self.cs_type.give(v, suffix) for k, v in value.items()}

    def each_get_admin_preview_object(self, value, *args, **kwargs):
        return (
            "{"
            + ",".join(
                [
                    f"<div class='subitem'><i>{k}</i>: {self.cs_type.get_admin_preview_object(v, *args, **kwargs)}</div>"
                    for k, v in value.items()
                ]
            )
            + "}"
        )


class EACH_SUFFIX(Enum):
    USE_OWN = auto()
    USE_PARENT = auto()
    SPLIT_OWN = auto()
    SPLIT_PARENT = auto()


class EachMixin:
    """
    Attributes:

    - `each` - the type of the subvalues.
    - `each_suffix_use` - how to use the suffixes. Can be `USE_OWN`, `USE_PARENT`, `SPLIT_OWN`, `SPLIT_PARENT`
    - `each_suffix_splitter` - the string that separates the suffixes. Applicable only when `each_suffix_use` is `SPLIT_OWN` or `SPLIT_PARENT`
    """

    admin_preview_as: PREVIEW = PREVIEW.HTML
    each: BaseEach = None
    each_suffix_use: EACH_SUFFIX = EACH_SUFFIX.USE_OWN
    each_suffix_splitter: str = "_by_"

    def process_each_validate(self, value):
        self.each.validate(value)

    def process_each_to_python(self, value):
        return self.each.to_python(value)

    def validate(self, value):
        super().validate(value)
        if value is None:
            return
        self.process_each_validate(value)

    def to_python(self, value):
        value = super().to_python(value)
        if value is None:
            return None
        return self.process_each_to_python(value)

    def get_each_suffix_splitter(self):
        return self.each_suffix_splitter

    def get_each_suffix_use(self):
        return self.each_suffix_use

    def get_help_format(self):
        yield from super().get_help_format()
        yield from self.each.get_help_format()

    def give_siffixes(self, value):
        if value is None:
            return None, None

        if self.get_each_suffix_use() == EACH_SUFFIX.USE_OWN:
            return value, None

        if self.get_each_suffix_use() == EACH_SUFFIX.USE_PARENT:
            return None, value

        splitter = self.get_each_suffix_splitter()
        if splitter not in value:
            if self.get_each_suffix_use() == EACH_SUFFIX.SPLIT_OWN:
                return value, None
            else:
                return None, value

        suffix, parent_suffix = value.split(splitter, 1)
        suffix = suffix
        return suffix, parent_suffix

    def give(self, value, suffix=None):
        suffix, each_suffix = self.give_siffixes(suffix)
        value = self.each.give(value, each_suffix)
        return super().give(value, suffix)

    def give_python_to_admin(self, value, *args, **kwargs):
        value = super().to_python(value)
        return self.each.give_python_to_admin(value, *args, **kwargs)

    def get_admin_preview_object(self, *args, **kwargs) -> str:
        return self.each.get_admin_preview_object(*args, **kwargs)

    def get_admin_head_css(self):
        return super().get_admin_head_css() + self.each.get_admin_head_css()

    def get_admin_head_js(self):
        return super().get_admin_head_js() + self.each.get_admin_head_js()

    def get_admin_head_css_raw(self):
        return super().get_admin_head_css_raw() + self.each.get_admin_head_css_raw()

    def get_admin_head_js_raw(self):
        return super().get_admin_head_js_raw() + self.each.get_admin_head_js_raw()



"""
A set of available utilites
"""

from typing import Iterator, Tuple, Type, Any, Callable, Union

import inspect
from importlib import import_module

from content_settings.types import BaseSetting, TCallableStr


def classes(setting_cls: Type[BaseSetting]) -> Iterator[Type[BaseSetting]]:
    """
    Returns an iterator of classes that are subclasses of the given class.
    """
    for cls in inspect.getmro(setting_cls):

        if not cls.__name__ or not cls.__module__:
            continue
        if cls.__module__ == "builtins":
            continue
        if (cls.__module__, cls.__name__) in (
            ("content_settings.types", "BaseSetting"),
            ("content_settings.types.basic", "SimpleString"),
        ) and cls != setting_cls:
            continue
        yield cls


def class_names(setting_cls: Type[BaseSetting]) -> Iterator[Tuple[str, str]]:
    """
    Returns an iterator of tuple with module and class name that are subclasses of the given class.
    """
    for cls in classes(setting_cls):
        yield (cls.__module__, cls.__name__)


def import_object(path: str) -> Any:
    """
    getting an object from the module by the path. `full.path.to.Object` -> `Object`
    """
    try:
        return import_module(path)
    except ImportError:
        parts = path.split(".")
        module = import_module(".".join(parts[:-1]))
        return getattr(module, parts[-1])


def function_has_argument(func: Callable, arg: str) -> bool:
    """
    Check if the function has the given argument in its definition.
    """
    return arg in inspect.signature(func).parameters


def is_bline(func: TCallableStr) -> bool:
    """
    Check if it is a string defined as b"string", or is other words bites string.

    The function is part of the future idea https://github.com/occipital/django-content-settings/issues/110
    """
    return isinstance(func, bytes)


def obj_base_str(obj: Any, call_base: Any = None) -> Callable:
    """
    if a given obj is not String - return the obj. If it is string than try to find it using call_base
    """
    if is_bline(obj):
        obj = obj.decode("utf-8")

    if not isinstance(obj, str):
        return obj

    if "." in obj:
        return import_object(obj)

    assert (
        call_base
    ), "the first argument can not be empty if the first argument a name of the function"
    if isinstance(call_base, str):
        call_base = import_object(call_base)

    return getattr(call_base, obj)


def call_base_str(func: TCallableStr, *args, call_base: Any = None, **kwargs) -> Any:
    """
    The goal of the function is to extend interface of callable attributes, so instead of passing a function you can pass a name of the function or full import path to the function.

    It is not only minimise the amout of import lines but also allows to use string attributes in `CONTENT_SETTINGS_DEFAULTS`.
    """

    func = obj_base_str(func, call_base)

    if function_has_argument(func, "call_base"):
        kwargs["call_base"] = call_base

    return func(*args, **kwargs)



"""
A set of functions that can be used inside migrations.
"""

from typing import Union, Optional, Type, Dict, Any
import json

from django.db import migrations, models
from django.contrib.auth.models import User


def RunImport(data: Union[str, dict], reverse_data: Optional[Union[str, dict]] = None):
    """
    The function that can be used inside your migration file.

    Args:
        data (Union[str, dict]): The data to import. Can be either a JSON string or a dictionary.
        reverse_data (Optional[Union[str, dict]], optional): The data to use for reversing the migration.
            Can be either a JSON string or a dictionary. If None, the migration will not be reversible. Defaults to None.

    Returns:
        migrations.RunPython: A RunPython operation that can be used in a migration file.

    Example:
        In your migration file:

        from content_settings.migrate import RunImport

        class Migration(migrations.Migration):
            dependencies = [
                ('content_settings', '0004_userdefined_preview'),
            ]

            operations = [
                RunImport({
                    "settings": {
                        "AFTER_TITLE": {
                            "value": "Best Settings Framework",
                            "version": ""
                        },
                        "ARTIST_LINE": {
                            "value": "",
                            "version": ""
                        },
                        "DAYS_WITHOUT_FAIL": {
                            "value": "5",
                            "version": "
                        },
                        "WEE": {
                            "value": "12",
                            "tags": "",
                            "help": "12",
                            "version": "",
                            "user_defined_type": "text"
                        }
                    }
                }),
            ]

    Note:
        This function is designed to be used within Django migration files. It handles the import of content settings,
        creating or updating them as necessary. If reverse_data is provided, it also sets up the reverse operation
        for the migration.
    """
    if isinstance(data, str):
        data = json.loads(data)

    def run(apps, schema_editor):

        import_settings(
            data,
            model_cs=apps.get_model("content_settings", "ContentSetting"),
            model_cs_history=apps.get_model(
                "content_settings", "HistoryContentSetting"
            ),
        )

    if reverse_data is None:
        return migrations.RunPython(run, lambda apps, schema_editor: None)

    if isinstance(reverse_data, str):
        reverse_data = json.loads(reverse_data)

    def reverse_run(apps, schema_editor):
        import_settings(
            reverse_data,
            model_cs=apps.get_model("content_settings", "ContentSetting"),
            model_cs_history=apps.get_model(
                "content_settings", "HistoryContentSetting"
            ),
        )

    return migrations.RunPython(run, reverse_run)


def import_settings(
    data: Dict[str, Any],
    model_cs: Type[models.Model],
    model_cs_history: Optional[Type[models.Model]] = None,
    user: Optional[User] = None,
) -> None:
    """
    Import content settings from a dictionary.

    Args:
        data (Dict[str, Any]): A dictionary containing the settings to import.
        model_cs (Type[models.Model]): The ContentSetting model class.
        model_cs_history (Optional[Type[models.Model]]): The ContentSettingHistory model class, if history tracking is enabled.
        user (Optional[User]): The user performing the import, if applicable.

    Returns:
        None

    This function iterates through the settings in the provided data dictionary,
    creating or updating ContentSetting objects as necessary. It also handles
    history tracking if a history model is provided.
    """
    for name, value in data["settings"].items():
        cs = model_cs.objects.filter(name=name).first()
        was_changed: Optional[bool] = cs is not None
        if value is None:
            was_changed = None

        if value is None:
            if cs:
                cs.delete()
        else:
            if not cs:
                cs = model_cs(name=name)
            for key, in_value in value.items():
                setattr(cs, key, in_value)
            cs.save()

        if user is not None:
            model_cs_history.update_last_record_for_name(name, user)

        elif model_cs_history is not None:
            history = model_cs_history(
                name=name, was_changed=was_changed, by_user=False
            )

            for key, in_value in value.items():
                setattr(history, key, in_value)
            history.save()



"""
The module contains types of different formats such as JSON, YAML, CSV, and so on.
"""

from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _

from typing import List, Optional, Tuple, Dict, Union

from .basic import SimpleText, PREVIEW, SimpleString
from .each import EachMixin, Keys, Item
from .mixins import EmptyNoneMixin
from . import optional, BaseSetting


# TODO: should have Empty None as well as JSON
class SimpleYAML(SimpleText):
    """
    YAML content settings type. Requires yaml module.
    """

    admin_preview_as = PREVIEW.PYTHON
    yaml_loader = None

    def get_help_format(self):
        return _(
            "Simple <a href='https://en.wikipedia.org/wiki/YAML' target='_blank'>YAML format</a>"
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            import yaml
        except ImportError:
            raise AssertionError("Please install pyyaml package")

    def get_yaml_loader(self):
        if self.yaml_loader is None:
            try:
                from yaml import CLoader as Loader
            except ImportError:
                from yaml import Loader
            return Loader
        return self.yaml_loader

    def to_python(self, value):
        value = super().to_python(value)
        if value is None:
            return None

        from yaml import load

        try:
            return load(value, Loader=self.get_yaml_loader())
        except Exception as e:
            raise ValidationError(str(e))


class SimpleJSON(EmptyNoneMixin, SimpleText):
    """
    JSON content settings type.
    """

    admin_preview_as = PREVIEW.PYTHON
    decoder_cls = None

    def get_help_format(self):
        return _(
            "Simple <a href='https://en.wikipedia.org/wiki/JSON' target='_blank'>JSON format</a>"
        )

    def get_decoder_cls(self):
        return self.decoder_cls

    def to_python(self, value):
        value = super().to_python(value)
        if value is None:
            return None
        from json import loads

        try:
            return loads(value, cls=self.get_decoder_cls())
        except Exception as e:
            raise ValidationError(str(e))


class SimpleRawCSV(SimpleText):
    """
    Type that converts simple CSV to list of lists.
    """

    admin_preview_as = PREVIEW.PYTHON

    csv_dialect = "unix"

    def get_help_format(self):
        return _(
            "Simple <a href='https://en.wikipedia.org/wiki/Comma-separated_values' target='_blank'>CSV format</a>"
        )

    def get_csv_reader(self, value):
        value = super().to_python(value)
        if value is None:
            return None

        from csv import reader
        from io import StringIO

        try:
            return reader(StringIO(value), dialect=self.csv_dialect)
        except Exception as e:
            raise ValidationError(str(e))

    def to_python(self, value):
        reader = self.get_csv_reader(value)
        return [list(row) for row in reader if row]


class KeysFromList(Keys):
    def to_python(self, value):
        return super().to_python({k: v for k, v in zip(self.cs_types.keys(), value)})

    def give_python_to_admin(self, value, name, **kwargs):
        return {
            k: cs.give_python_to_admin(v, name, **kwargs)
            for (k, cs), v in zip(self.cs_types.items(), value)
        }


class KeysFromListByList(KeysFromList):
    def __init__(self, cs_type, keys):
        super().__init__(**{k: cs_type for k in keys})


class SimpleCSV(EachMixin, SimpleRawCSV):
    """
    Type that converts simple CSV to list of dictionaries.

    Attributes:

    - `csv_fields` (dict, tuple or list): defines the structure of the CSV. The structure definition used by `EachMixin`
    - `csv_fields_list_type` (BaseSetting): the type of the list elements in the `csv_fields` if it is not dict.

    Examples:

    ```python
    SimpleCSV(csv_fields=["name", "price"])
    SimpleCSV(csv_fields={"name": SimpleString(), "price": SimpleDecimal()})
    SimpleCSV(csv_fields=["name", "price"], csv_fields_list_type=SimpleString())
    ```
    """

    csv_fields: Optional[Union[List, Tuple, Dict]] = None
    csv_fields_list_type: BaseSetting = SimpleString(optional)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert self.csv_fields is not None, "csv_fields cannot be None"
        assert isinstance(
            self.csv_fields, (list, tuple, dict)
        ), "csv_fields must be list, tuple or dict"

        self.each = Item(
            KeysFromList(**self.csv_fields)
            if isinstance(self.csv_fields, dict)
            else KeysFromListByList(self.csv_fields_list_type, self.csv_fields)
        )



"""
Available middlewares for the content settings.
"""

from django.urls import reverse

from .context_managers import content_settings_context
from .models import UserPreview
from .settings import PREVIEW_ON_SITE_SHOW


def preivew_on_site(get_response):
    """
    the middleware required for previewing the content settings on the site.

    It checks content_settings.can_preview_on_site permission for the user and if the user has it, then the middleware will preview the content settings for the user.
    """

    def middleware(request):
        if not PREVIEW_ON_SITE_SHOW or not request.user.has_perm(
            "content_settings.can_preview_on_site"
        ):
            return get_response(request)

        if request.path.startswith(reverse("admin:index") + "content_settings/"):
            return get_response(request)

        preview_settings = UserPreview.get_context_dict(request.user)
        if not preview_settings:
            return get_response(request)

        with content_settings_context(**preview_settings, _raise_errors=False):
            return get_response(request)

    return middleware



"""
Widgets that can be used in `widget` attribute of the content settings.

It includes all of the widgets from django.forms.widgets + several custom widgets.
"""

from django.forms.widgets import *


class LongInputMix:
    """
    Mixin that makes input maximum long
    """

    def __init__(self, attrs=None):
        attrs = attrs or {}
        style = attrs.pop("style", "")
        style = "width: calc(100% - 14px);" + style
        attrs["style"] = style
        super().__init__(attrs=attrs)


class LongTextInput(LongInputMix, TextInput):
    """
    Long text input
    """

    pass


class LongTextarea(LongInputMix, Textarea):
    """
    Long textarea
    """

    pass


class LongURLInput(LongInputMix, URLInput):
    """
    Long URL Input
    """


class LongEmailInput(LongInputMix, EmailInput):
    """
    Long Email Input
    """

    pass



import argparse
import json
from django.core.management.base import BaseCommand
from content_settings.export import preview_data, import_to
from django.contrib.auth import get_user_model

User = get_user_model()


class Command(BaseCommand):
    help = "Import content settings from a JSON file."

    def add_arguments(self, parser):
        parser.add_argument(
            "filename",
            type=str,
            help="Path to the JSON file containing content settings.",
        )
        parser.add_argument(
            "--import",
            action="store_true",
            default=False,
            help="Apply the import if set, otherwise just preview.",
        )
        parser.add_argument(
            "--preview-for",
            type=str,
            default=None,
            help="Import as a preview for a specific user.",
        )
        parser.add_argument(
            "--names",
            nargs="+",
            type=str,
            help="Names of specific content settings to import.",
        )

        parser.add_argument(
            "--show-skipped",
            action="store_true",
            default=False,
            help="Show skipped settings.",
        )
        parser.add_argument(
            "--show-only-errors",
            action="store_true",
            default=False,
            help="Show only errors.",
        )

    def handle(self, *args, **options):
        filename = options["filename"]
        do_import = options["import"]
        preview_for = options["preview_for"]
        names = options["names"]
        show_skipped = options["show_skipped"]
        show_only_errors = options["show_only_errors"]

        with open(filename, "r") as file:
            data = json.load(file)

        if preview_for:
            preview_for = User.objects.get(username=preview_for)

        errors, applied, skipped = preview_data(data, preview_for)

        if skipped and show_skipped and not show_only_errors:
            self.stdout.write("Skipped:")
            self.stdout.write(json.dumps(skipped, indent=2))

        if errors:
            self.stderr.write("Errors:")
            self.stderr.write(json.dumps(errors, indent=2))

        if applied and not show_only_errors:
            self.stdout.write("Applied:")
            self.stdout.write(json.dumps(applied, indent=2))

        if do_import or preview_for:
            if names:
                applied = [row for row in applied if row["name"] in names]

            import_to(data, applied, preview_for is not None, preview_for)

            if preview_for:
                self.stdout.write(
                    self.style.SUCCESS(
                        "Preview saved for user %s" % preview_for.username
                    )
                )
            else:
                self.stdout.write(self.style.SUCCESS("Import completed."))



"""
Classes that uses for lazy loading of objects.

Check `BaseString.lazy_give` and `conf.lazy_prefix`
"""

import operator


class LazyObject:
    def __init__(self, factory):
        # Assign using __dict__ to avoid the setattr method.
        self.__dict__["_factory"] = factory

    def new_method_proxy(func):
        """
        Util function to help us route functions
        to the nested object.
        """

        def inner(self, *args):
            return func(self._factory(), *args)

        return inner

    def __call__(self, *args, **kwargs):
        return self._factory()(*args, **kwargs)

    __getattr__ = new_method_proxy(getattr)
    __bytes__ = new_method_proxy(bytes)
    __str__ = new_method_proxy(str)
    __bool__ = new_method_proxy(bool)
    __dir__ = new_method_proxy(dir)
    __hash__ = new_method_proxy(hash)
    __class__ = property(new_method_proxy(operator.attrgetter("__class__")))
    __eq__ = new_method_proxy(operator.eq)
    __lt__ = new_method_proxy(operator.lt)
    __le__ = new_method_proxy(operator.le)
    __gt__ = new_method_proxy(operator.gt)
    __ge__ = new_method_proxy(operator.ge)
    __ne__ = new_method_proxy(operator.ne)
    __mod__ = new_method_proxy(operator.mod)
    __hash__ = new_method_proxy(hash)
    __getitem__ = new_method_proxy(operator.getitem)
    __iter__ = new_method_proxy(iter)
    __len__ = new_method_proxy(len)
    __contains__ = new_method_proxy(operator.contains)
    __mul__ = new_method_proxy(operator.mul)
    __rmul__ = new_method_proxy(operator.mul)
    __not__ = new_method_proxy(operator.not_)
    __bool__ = new_method_proxy(bool)
    __len__ = new_method_proxy(len)
    __abs__ = new_method_proxy(operator.abs)
    __add__ = new_method_proxy(operator.add)
    __radd__ = new_method_proxy(operator.add)
    __and__ = new_method_proxy(operator.and_)
    __floordiv__ = new_method_proxy(operator.floordiv)
    __index__ = new_method_proxy(operator.index)
    __inv__ = new_method_proxy(operator.inv)
    __invert__ = new_method_proxy(operator.invert)
    __mod__ = new_method_proxy(operator.mod)
    __matmul__ = new_method_proxy(operator.matmul)
    __neg__ = new_method_proxy(operator.neg)
    __or__ = new_method_proxy(operator.or_)
    __pos__ = new_method_proxy(operator.pos)
    __pow__ = new_method_proxy(operator.pow)
    __sub__ = new_method_proxy(operator.sub)
    __truediv__ = new_method_proxy(operator.truediv)
    __xor__ = new_method_proxy(operator.xor)
    __concat__ = new_method_proxy(operator.concat)
    __contains__ = new_method_proxy(operator.contains)
    __getitem__ = new_method_proxy(operator.getitem)
    __iadd__ = new_method_proxy(operator.iadd)
    __iand__ = new_method_proxy(operator.iand)
    __iconcat__ = new_method_proxy(operator.iconcat)
    __ifloordiv__ = new_method_proxy(operator.ifloordiv)
    __ilshift__ = new_method_proxy(operator.ilshift)
    __imod__ = new_method_proxy(operator.imod)
    __imul__ = new_method_proxy(operator.imul)
    __imatmul__ = new_method_proxy(operator.imatmul)
    __ior__ = new_method_proxy(operator.ior)
    __ipow__ = new_method_proxy(operator.ipow)
    __isub__ = new_method_proxy(operator.isub)
    __itruediv__ = new_method_proxy(operator.itruediv)
    __ixor__ = new_method_proxy(operator.ixor)
    __next__ = new_method_proxy(next)
    __reversed__ = new_method_proxy(reversed)
    __round__ = new_method_proxy(round)



"""
Those are the views can be used in the Integration with the Project.
"""

from django.http import HttpResponseNotFound, HttpResponseForbidden, HttpResponse
from django.views.generic import View
from django.utils.translation import gettext as _
from content_settings.conf import content_settings

from .conf import ALL, split_attr, content_settings
from .caching import get_type_by_name


def gen_startswith(startswith: str):
    """
    for names attribute of FetchSettingsView, to find settings by name starts with `startswith`
    """

    def _(request):
        for name in dir(content_settings):
            if name.startswith(startswith):
                yield name

    return _


def gen_hastag(tag: str):
    """
    for names attribute of FetchSettingsView, to find settings by tag
    """

    def _(request):
        for name in dir(content_settings):
            if tag in get_type_by_name(name).get_tags():
                yield name

    return _


def gen_all():
    """
    for names attribute of FetchSettingsView, to find all settings
    """

    def _(request):
        for name in dir(content_settings):
            yield name

    return _


class FetchSettingsView(View):
    """
    A View for featching settings from the content settings.

    Use attribute `names` to define the names of the settings to fetch.

    ```
    FetchSettingsView.as_view(
        names=[
            "DESCRIPTION",
            "OPEN_DATE",
            "TITLE",
        ]
    )
    ```

    Suffix can be used in names

    ```
    FetchSettingsView.as_view(
        names=[
            "TITLE",
            "BOOKS__available_names",
        ]
    ),
    ```

    function for getting names by specific conditions can be used

    ```
    FetchSettingsView.as_view(names=gen_hastag("general")),
    ```

    or combinations of them

    ```
    FetchSettingsView.as_view(names=(gen_startswith("IS_"), "TITLE")),
    ```

    """

    names = ()
    show_error_headers = True

    def get_names(self, request):
        if callable(self.names):
            yield from self.names(request)
        else:
            for value in self.names:
                if callable(value):
                    yield from value(request)
                else:
                    yield value

    def get(self, request):
        ret = []
        errors = []
        for val in self.get_names(request):
            if isinstance(val, tuple):
                key, val = val
            else:
                key = val

            prefix, name, suffix = split_attr(val)
            assert prefix is None, "prefix is not None"

            cs_type = get_type_by_name(name)
            if cs_type is None:
                errors.append(f"{key}: not found")
                continue

            if not cs_type.can_fetch(request.user):
                errors.append(f"{key}: permission denied")
                continue

            value = getattr(content_settings, val)
            ret.append(
                (
                    key,
                    cs_type.json_view_value(
                        value, suffix=suffix, request=request, name=name
                    ),
                )
            )

        return HttpResponse(
            "{" + (",".join(f'"{name}":{value}' for name, value in ret)) + "}",
            content_type="application/json",
            headers=(
                {"X-Content-Settings-Errors": ";".join(errors)}
                if errors and self.show_error_headers
                else {}
            ),
        )


class FetchAllSettingsView(FetchSettingsView):
    names = staticmethod(gen_all())



# Generated by Django 4.2.13 on 2024-07-29 10:46

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("content_settings", "0003_user_preview"),
    ]

    operations = [
        migrations.AddField(
            model_name="userpreview",
            name="from_help",
            field=models.TextField(null=True, verbose_name="From Help"),
        ),
        migrations.AddField(
            model_name="userpreview",
            name="from_tags",
            field=models.TextField(null=True, verbose_name="From Tags"),
        ),
        migrations.AddField(
            model_name="userpreview",
            name="from_user_defined_type",
            field=models.CharField(
                max_length=50, null=True, verbose_name="From User Defined Type"
            ),
        ),
        migrations.AddField(
            model_name="userpreview",
            name="help",
            field=models.TextField(null=True, verbose_name="Help"),
        ),
        migrations.AddField(
            model_name="userpreview",
            name="tags",
            field=models.TextField(null=True, verbose_name="Tags"),
        ),
        migrations.AddField(
            model_name="userpreview",
            name="user_defined_type",
            field=models.CharField(
                max_length=50, null=True, verbose_name="User Defined Type"
            ),
        ),
        migrations.AddField(
            model_name="userpreviewhistory",
            name="help",
            field=models.TextField(null=True),
        ),
        migrations.AddField(
            model_name="userpreviewhistory",
            name="tags",
            field=models.TextField(null=True),
        ),
        migrations.AddField(
            model_name="userpreviewhistory",
            name="user_defined_type",
            field=models.CharField(max_length=50, null=True),
        ),
        migrations.AlterField(
            model_name="contentsetting",
            name="help",
            field=models.TextField(blank=True, null=True, verbose_name="Help"),
        ),
        migrations.AlterField(
            model_name="contentsetting",
            name="name",
            field=models.CharField(
                max_length=200,
                unique=True,
                validators=[
                    django.core.validators.RegexValidator(
                        message="Value must be uppercase, can include numbers and underscores, and cannot start with a number.",
                        regex="^[A-Z][A-Z0-9_]*$",
                    )
                ],
                verbose_name="Name",
            ),
        ),
        migrations.AlterField(
            model_name="contentsetting",
            name="tags",
            field=models.TextField(blank=True, null=True, verbose_name="Tags"),
        ),
        migrations.AlterField(
            model_name="contentsetting",
            name="user_defined_type",
            field=models.CharField(
                db_index=True,
                default=None,
                max_length=50,
                null=True,
                verbose_name="User Defined Type",
            ),
        ),
        migrations.AlterField(
            model_name="contentsetting",
            name="value",
            field=models.TextField(blank=True, verbose_name="Value"),
        ),
        migrations.AlterField(
            model_name="contentsetting",
            name="version",
            field=models.CharField(max_length=50, null=True, verbose_name="Version"),
        ),
        migrations.AlterField(
            model_name="historycontentsetting",
            name="by_user",
            field=models.BooleanField(
                choices=[(True, "by user"), (False, "by app"), (None, "unknown")],
                default=None,
                null=True,
                verbose_name="By User",
            ),
        ),
        migrations.AlterField(
            model_name="historycontentsetting",
            name="created_on",
            field=models.DateTimeField(
                default=django.utils.timezone.now, verbose_name="Created On"
            ),
        ),
        migrations.AlterField(
            model_name="historycontentsetting",
            name="help",
            field=models.TextField(blank=True, null=True, verbose_name="Help"),
        ),
        migrations.AlterField(
            model_name="historycontentsetting",
            name="name",
            field=models.CharField(max_length=200, verbose_name="Name"),
        ),
        migrations.AlterField(
            model_name="historycontentsetting",
            name="tags",
            field=models.TextField(blank=True, null=True, verbose_name="Tags"),
        ),
        migrations.AlterField(
            model_name="historycontentsetting",
            name="user",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.AUTH_USER_MODEL,
                verbose_name="User",
            ),
        ),
        migrations.AlterField(
            model_name="historycontentsetting",
            name="user_defined_type",
            field=models.CharField(
                default=None, max_length=50, null=True, verbose_name="User Defined Type"
            ),
        ),
        migrations.AlterField(
            model_name="historycontentsetting",
            name="value",
            field=models.TextField(blank=True, null=True, verbose_name="Value"),
        ),
        migrations.AlterField(
            model_name="historycontentsetting",
            name="version",
            field=models.CharField(max_length=50, null=True, verbose_name="Version"),
        ),
        migrations.AlterField(
            model_name="historycontentsetting",
            name="was_changed",
            field=models.BooleanField(
                choices=[(True, "Changed"), (False, "Added"), (None, "Removed")],
                default=True,
                null=True,
                verbose_name="Was Changed",
            ),
        ),
        migrations.AlterField(
            model_name="userpreview",
            name="from_value",
            field=models.TextField(null=True, verbose_name="From Value"),
        ),
        migrations.AlterField(
            model_name="userpreview",
            name="name",
            field=models.CharField(max_length=200, verbose_name="Name"),
        ),
        migrations.AlterField(
            model_name="userpreview",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
                verbose_name="User",
            ),
        ),
        migrations.AlterField(
            model_name="userpreview",
            name="value",
            field=models.TextField(verbose_name="Value"),
        ),
        migrations.AlterField(
            model_name="usertagsetting",
            name="name",
            field=models.CharField(max_length=200, verbose_name="Name"),
        ),
        migrations.AlterField(
            model_name="usertagsetting",
            name="tag",
            field=models.CharField(max_length=200, verbose_name="Tag"),
        ),
        migrations.AlterField(
            model_name="usertagsetting",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
                verbose_name="User",
            ),
        ),
        migrations.AddIndex(
            model_name="historycontentsetting",
            index=models.Index(
                fields=["created_on", "user_id", "by_user"], name="static_batch_idx"
            ),
        ),
    ]
